from pathlib import Path


import numpy as np
import xarray as xr


BASE_PATH = ''


def set_base_path(new_path):
    global BASE_PATH
    if new_path is not None:
        BASE_PATH = new_path


CF_MAPPING = {
    't2m': 'air_temperature',
    'si10': 'wind_speed',
    'dswrf': 'ghi',
    'vbdsf': 'dni',
    'vddsf': 'dhi',
    'tcdc': 'cloud_cover'
}


def _load_pnt(ds, latitude, longitude, limit):
    lats = ds.latitude.values
    lons = ds.longitude.values
    if lats.ndim == 1:
        # ds uses lat and lon as primary coordinates. We still want to use
        # tunnel_fast to determine distance so that we can check that
        # the query point is not too far out of the domain.
        # tunnel_fast requires 2D grid
        lons, lats = np.meshgrid(lons, lats)
    iy_min, ix_min = tunnel_fast(lats, lons, latitude, longitude,
                                 limit=limit)
    # could avoid this if statement if we only use positional indexing
    # like ds[iy_min, ix_min] but this seems safer
    if ds.latitude.ndim == 1:
        pnt = ds.isel(latitude=iy_min, longitude=ix_min)
    else:
        pnt = ds.isel(y=iy_min, x=ix_min)
    return pnt


def load_forecast(
        latitude, longitude, init_time, start, end, model,
        variables=('ghi', 'dni', 'dhi', 'air_temperature', 'wind_speed'),
        base_path=None):
    """Load NWP model data.

    Parameters
    ----------
    latitude : float

    longitude : float

    init_time : pd.Timestamp
        Full datetime of a model initialization

    start : pd.Timestamp

    end : pd.Timestamp

    model : str
        Name of model. Must be one of:

            * 'hrrr_hourly'
            * 'hrrr_subhourly'
            * 'gfs_0p25'
            * 'rap'
            * 'nam_12km'
            * 'gefs_p{num}' where num is '01' to '20', 'gefs_c00',
              'gefs_avg', or 'gefs_spr'

    variables : list of str
        The variables to load.

    Returns
    -------
    By default:

      * ghi : pd.Series
      * dni : pd.Series
      * dhi : pd.Series
      * air_temperature : pd.Series
      * wind_speed : pd.Series

    Raises
    ------
    ValueError : Raised if the requested variable is not found.
    """
    base_path = base_path if base_path is not None else BASE_PATH
    if 'gefs' in model:
        # account for slightly different file layout for gefs
        model_path = 'gefs'
    else:
        model_path = model
    filepath = (Path(base_path) / model_path /
                init_time.strftime('%Y/%m/%d/%H') / (model + '.nc'))
    if not filepath.is_file():
        raise FileNotFoundError(f'{filepath} does not exist')
    mapping_subset = {k: v for k, v in CF_MAPPING.items() if v in variables}

    limit = 500  # maximum distance from point to closest grid point
    # Time in file are stored unlocalized but are UTC
    # so convert localized start/end to UTC then drop tz
    if start.tzinfo is not None:
        start = start.tz_convert('UTC').tz_localize(None)
    if end.tzinfo is not None:
        end = end.tz_convert('UTC').tz_localize(None)
    with xr.open_dataset(filepath) as ds:
        pnt = _load_pnt(ds, latitude, longitude, limit)
        pnt = pnt.sel(time=slice(start, end))
        pnt = pnt.rename(mapping_subset)
        pnt['air_temperature'] -= 273.15  # convert Kelvin to deg C
        series = [pnt[variable].to_series().tz_localize('UTC')
                  for variable in variables]
        return series


def tunnel_fast(latvar, lonvar, lat0, lon0, limit=None):
    """
    Find closest point in a set of (lat, lon) points to specified point.

    Returns iy, ix such that the square of the tunnel distance
    between (latval[it, ix], lonval[iy, ix]) and (lat0, lon0)
    is minimum.

    Parameters
    ----------
    latvar : numpy.array
        Model latitudes.
    lonvar : numpy.array
        Model longitudes.
    lat0 : float
        Latitude of desired point.
    lon0 : float
        Longitude of desired point.
    limit : None or float
        Maximum distance allowed in units of km.

    Returns
    -------
    iy : int
        Index of point in latitude array that minimizes distance to
        desired point.
    ix : int
        Index of point in longitude array that minimizes distance to
        desired point.

    Raises
    ------
    ValueError
        If closest point exceeds minimum distance requirement.

    References
    ----------
    https://github.com/Unidata/python-workshop/blob/fall-2016/notebooks/netcdf-by-coordinates.ipynb
    """
    rad_factor = np.pi/180.0
    latvals = latvar * rad_factor
    lonvals = lonvar * rad_factor
    lat0_rad = lat0 * rad_factor
    lon0_rad = lon0 * rad_factor
    clat, clon = np.cos(latvals), np.cos(lonvals)
    slat, slon = np.sin(latvals), np.sin(lonvals)
    delX = np.cos(lat0_rad)*np.cos(lon0_rad) - clat*clon
    delY = np.cos(lat0_rad)*np.sin(lon0_rad) - clat*slon
    delZ = np.sin(lat0_rad) - slat
    dist_sq = delX**2 + delY**2 + delZ**2
    if limit is not None:
        dist = 6378.1 * np.sqrt(dist_sq.min())
        if dist > limit:
            raise ValueError('Maximum distance limit exceeded')
    minindex_1d = dist_sq.argmin()  # 1D index of minimum element
    iy_min, ix_min = np.unravel_index(minindex_1d, latvals.shape)
    return iy_min, ix_min



import logging


import pandas as pd
import pandas.testing as pdt
import pytest


from solarforecastarbiter import utils


def _make_aggobs(obsid, ef=pd.Timestamp('20191001T1100Z'),
                 eu=None, oda=None):
    return {
        'observation_id': obsid,
        'effective_from': ef,
        'effective_until': eu,
        'observation_deleted_at': oda
    }


nindex = pd.date_range(start='20191004T0000Z',
                       freq='1h', periods=10)


@pytest.fixture()
def ids():
    return ['f2844284-ea0a-11e9-a7da-f4939feddd82',
            'f3e310ba-ea0a-11e9-a7da-f4939feddd82',
            '09ed7cf6-ea0b-11e9-a7da-f4939feddd82',
            '0fe9f2ba-ea0b-11e9-a7da-f4939feddd82',
            '67ea9200-ea0e-11e9-832b-f4939feddd82']


@pytest.fixture()
def aggobs(ids):
    return tuple([
        _make_aggobs(ids[0]),
        _make_aggobs(ids[1], pd.Timestamp('20191004T0501Z')),
        _make_aggobs(ids[2], eu=pd.Timestamp('20191004T0400Z')),
        _make_aggobs(ids[2], pd.Timestamp('20191004T0700Z'),
                     eu=pd.Timestamp('20191004T0800Z')),
        _make_aggobs(ids[2], pd.Timestamp('20191004T0801Z')),
        _make_aggobs(ids[3], oda=pd.Timestamp('20191005T0000Z')),
        _make_aggobs(ids[4], oda=pd.Timestamp('20191009T0000Z'),
                     eu=pd.Timestamp('20191003T0000Z'))
    ])


def test_compute_aggregate(aggobs, ids):
    data = {id_: pd.DataFrame({'value': [1] * 10, 'quality_flag': [0] * 10},
                              index=nindex)
            for id_ in ids[:3]}
    agg = utils.compute_aggregate(data, '1h', 'ending',
                                  'UTC', 'sum', aggobs[:-2])
    pdt.assert_frame_equal(agg, pd.DataFrame(
        {'value': pd.Series([2.0, 2.0, 2.0, 2.0, 2.0, 1.0, 2.0, 3.0, 3.0, 3.0],
                            index=nindex),
         'quality_flag':  pd.Series([0]*10, index=nindex)})
        )


def test_compute_aggregate_missing_from_data(aggobs, ids):
    data = {id_: pd.DataFrame({'value': [1] * 10, 'quality_flag': [0] * 10},
                              index=nindex)
            for id_ in ids[:3]}
    aggobs = list(aggobs[:-2]) + [
        _make_aggobs('09ed7cf6-ea0b-11e9-a7da-f4939fed889')]
    with pytest.raises(KeyError):
        utils.compute_aggregate(data, '1h', 'ending',
                                'UTC', 'sum', aggobs)


def test_compute_aggregate_empty_data(aggobs, ids):
    data = {}
    with pytest.raises(KeyError):
        utils.compute_aggregate(data, '1h', 'ending',
                                'UTC', 'sum', aggobs[:2])


@pytest.mark.filterwarnings('ignore::UserWarning')
def test_compute_aggregate_missing_data(aggobs, ids):
    data = {id_: pd.DataFrame({'value': [1] * 10, 'quality_flag': [0] * 10},
                              index=nindex)
            for id_ in ids[:3]}
    data[ids[-1]] = pd.DataFrame({'value': [1] * 8, 'quality_flag': [0] * 8},
                                 index=nindex[:-2])
    aggobs = list(aggobs[:-2]) + [_make_aggobs(ids[-1])]
    agg = utils.compute_aggregate(data, '1h', 'ending',
                                  'UTC', 'sum', aggobs)
    pdt.assert_frame_equal(agg, pd.DataFrame(
        {'value': pd.Series(
            [3.0, 3.0, 3.0, 3.0, 3.0, 2.0, 3.0, 4.0, None, None],
            index=nindex),
         'quality_flag':  pd.Series([0]*10, index=nindex)})
        )


def test_compute_aggregate_deleted_not_removed(aggobs, ids):
    data = {id_: pd.DataFrame({'value': [1] * 10, 'quality_flag': [0] * 10},
                              index=nindex)
            for id_ in ids}
    with pytest.raises(ValueError):
        utils.compute_aggregate(data, '1h', 'ending',
                                'UTC', 'sum', aggobs)


def test_compute_aggregate_deleted_not_removed_yet(aggobs, ids):
    data = {id_: pd.DataFrame({'value': [1] * 10, 'quality_flag': [0] * 10},
                              index=nindex)
            for id_ in ids[:3]}
    # with last aggobs, would try and get data before effective_until,
    # but was deleted, so raise error
    aggobs = list(aggobs[:-2]) + [
        _make_aggobs(ids[4], oda=pd.Timestamp('20191009T0000Z'),
                     eu=pd.Timestamp('20191004T0700Z'))]
    with pytest.raises(ValueError):
        utils.compute_aggregate(data, '1h', 'ending',
                                'UTC', 'sum', aggobs)


def test_compute_aggregate_deleted_but_removed_before(aggobs, ids):
    data = {id_: pd.DataFrame({'value': [1] * 10, 'quality_flag': [0] * 10},
                              index=nindex)
            for id_ in ids[:3]}
    # aggobs[-1] properly removed
    aggobs = list(aggobs[:-2]) + [aggobs[-1]]
    agg = utils.compute_aggregate(data, '1h', 'ending',
                                  'UTC', 'sum', aggobs)
    pdt.assert_frame_equal(agg, pd.DataFrame(
        {'value': pd.Series([2.0, 2.0, 2.0, 2.0, 2.0, 1.0, 2.0, 3.0, 3.0, 3.0],
                            index=nindex),
         'quality_flag':  pd.Series([0]*10, index=nindex)}))


def test_compute_aggregate_mean(aggobs, ids):
    data = {id_: pd.DataFrame({'value': [1] * 10, 'quality_flag': [0] * 10},
                              index=nindex)
            for id_ in ids[:3]}
    agg = utils.compute_aggregate(data, '1h', 'ending',
                                  'UTC', 'mean', aggobs[:-2])
    pdt.assert_frame_equal(agg, pd.DataFrame(
        {'value': pd.Series([1.0] * 10, index=nindex),
         'quality_flag':  pd.Series([0]*10, index=nindex)})
        )


@pytest.mark.filterwarnings('ignore::UserWarning')
def test_compute_aggregate_no_overlap(ids):
    data = {ids[0]: pd.DataFrame(
        {'value': [1, 2, 3], 'quality_flag': [2, 10, 338]},
        index=pd.DatetimeIndex([
            '20191002T0100Z', '20191002T0130Z', '20191002T0230Z'])),
            ids[1]: pd.DataFrame(
        {'value': [3, 2, 1], 'quality_flag': [9, 880, 10]},
        index=pd.DatetimeIndex([
            '20191002T0200Z', '20191002T0230Z', '20191002T0300Z']))}
    aggobs = [_make_aggobs(ids[0]),
              _make_aggobs(ids[1], pd.Timestamp('20191002T0200Z'))]
    agg = utils.compute_aggregate(data, '30min', 'ending',
                                  'UTC', 'median', aggobs)
    expected = pd.DataFrame(
        {'value': [1.0, 2.0, None, 2.5, None],
         'quality_flag': [2, 10, 9, 338 | 880, 10]},
        index=pd.DatetimeIndex([
            '20191002T0100Z', '20191002T0130Z', '20191002T0200Z',
            '20191002T0230Z', '20191002T0300Z']))
    pdt.assert_frame_equal(agg, expected)


def test_compute_aggregate_missing_before_effective(ids):
    data = {ids[0]: pd.DataFrame(
        {'value': [1, 2, 3, 0, 0], 'quality_flag': [2, 10, 338, 0, 0]},
        index=pd.DatetimeIndex([
            '20191002T0100Z', '20191002T0130Z', '20191002T0200Z',
            '20191002T0230Z', '20191002T0300Z'])),
            ids[1]: pd.DataFrame(
        {'value': [None, 2.0, 1.0], 'quality_flag': [0, 880, 10]},
        index=pd.DatetimeIndex([
            '20191002T0200Z', '20191002T0230Z', '20191002T0300Z']))}
    aggobs = [_make_aggobs(ids[0]),
              _make_aggobs(ids[1], pd.Timestamp('20191002T0201Z'))]
    agg = utils.compute_aggregate(data, '30min', 'ending',
                                  'UTC', 'max', aggobs)
    expected = pd.DataFrame(
        {'value': [1.0, 2.0, 3.0, 2.0, 1.0],
         'quality_flag': [2, 10, 338, 880, 10]},
        index=pd.DatetimeIndex([
            '20191002T0100Z', '20191002T0130Z', '20191002T0200Z',
            '20191002T0230Z', '20191002T0300Z']))
    pdt.assert_frame_equal(agg, expected)


def test_compute_aggregate_bad_cols():
    data = {'a': pd.DataFrame([0], index=pd.DatetimeIndex(
        ['20191001T1200Z']))}
    with pytest.raises(KeyError):
        utils.compute_aggregate(data, '1h', 'ending', 'UTC',
                                'mean', [_make_aggobs('a')])


def test__observation_valid(aggobs):
    out = utils._observation_valid(
        nindex, 'f2844284-ea0a-11e9-a7da-f4939feddd82', aggobs)
    pdt.assert_series_equal(out, pd.Series(True, index=nindex))


def test__observation_valid_ended(aggobs):
    out = utils._observation_valid(
        nindex, 'f3e310ba-ea0a-11e9-a7da-f4939feddd82', aggobs)
    pdt.assert_series_equal(out, pd.Series([False] * 6 + [True] * 4,
                                           index=nindex))


def test__observation_valid_many(aggobs):
    out = utils._observation_valid(
        nindex, '09ed7cf6-ea0b-11e9-a7da-f4939feddd82', aggobs)
    pdt.assert_series_equal(out, pd.Series(
        [True, True, True, True, True, False, False, True, True, True],
        index=nindex))


def test__observation_valid_deleted(aggobs):
    with pytest.raises(ValueError):
        utils._observation_valid(
            nindex, '0fe9f2ba-ea0b-11e9-a7da-f4939feddd82', aggobs)


def test__observation_valid_deleted_before(aggobs):
    out = utils._observation_valid(
        nindex, '67ea9200-ea0e-11e9-832b-f4939feddd82', aggobs)
    pdt.assert_series_equal(out, pd.Series(False, index=nindex))


@pytest.mark.parametrize('length,label,expected', [
    ('15min', 'ending', pd.date_range(start='20191004T0700Z',
                                      end='20191004T0745Z',
                                      freq='15min')),
    ('15min', 'beginning', pd.date_range(
        start='20191004T0700Z', end='20191004T0745Z',
        freq='15min')),
    ('1h', 'ending', pd.DatetimeIndex(['20191004T0700Z', '20191004T0800Z'])),
    ('1h', 'beginning', pd.DatetimeIndex(['20191004T0700Z'])),
    ('20min', 'ending', pd.DatetimeIndex([
        '20191004T0700Z', '20191004T0720Z', '20191004T0740Z',
        '20191004T0800Z'])),
    ('20min', 'beginning', pd.DatetimeIndex([
        '20191004T0700Z', '20191004T0720Z', '20191004T0740Z'])),
])
def test__make_aggregate_index(length, label, expected):
    test_data = {
        0: pd.DataFrame(range(5), index=pd.date_range(
            '20191004T0700Z', freq='7min', periods=5)),  # end 35
        1: pd.DataFrame(range(4), index=pd.date_range(
            '20191004T0015-0700', freq='10min', periods=4))}  # end 45
    out = utils._make_aggregate_index(test_data, length, label, 'UTC')
    pdt.assert_index_equal(out, expected)


@pytest.mark.parametrize('length,label,expected', [
    ('15min', 'ending', pd.date_range(start='20191004T0715Z',
                                      end='20191004T0745Z',
                                      freq='15min')),
    ('15min', 'beginning', pd.date_range(
        start='20191004T0700Z', end='20191004T0730Z',
        freq='15min')),
    ('1h', 'ending', pd.DatetimeIndex(['20191004T0800Z'])),
    ('1h', 'beginning', pd.DatetimeIndex(['20191004T0700Z'])),
    ('20min', 'ending', pd.DatetimeIndex([
        '20191004T0720Z', '20191004T0740Z'])),
    ('20min', 'beginning', pd.DatetimeIndex([
        '20191004T0700Z', '20191004T0720Z'])),
])
def test__make_aggregate_index_offset_right(length, label, expected):
    test_data = {
        0: pd.DataFrame(range(6), index=pd.date_range(
            '20191004T0701Z', freq='7min', periods=6))  # end 35
        }
    out = utils._make_aggregate_index(test_data, length, label, 'UTC')
    pdt.assert_index_equal(out, expected)


@pytest.mark.parametrize('length,label,expected', [
    ('15min', 'ending', pd.date_range(start='20191004T0700Z',
                                      end='20191004T0745Z',
                                      freq='15min')),
    ('15min', 'beginning', pd.date_range(
        start='20191004T0645Z', end='20191004T0730Z',
        freq='15min')),
    ('1h', 'ending', pd.DatetimeIndex(['20191004T0700Z', '20191004T0800Z'])),
    ('1h', 'beginning', pd.DatetimeIndex(['20191004T0600Z',
                                          '20191004T0700Z'])),
    ('20min', 'ending', pd.DatetimeIndex([
        '20191004T0700Z', '20191004T0720Z', '20191004T0740Z'])),
    ('20min', 'beginning', pd.DatetimeIndex([
        '20191004T0640Z', '20191004T0700Z', '20191004T0720Z'])),
    ('36min', 'ending', pd.DatetimeIndex(['20191004T0712Z',
                                          '20191004T0748Z'])),
    ('36min', 'beginning', pd.DatetimeIndex(['20191004T0636Z',
                                             '20191004T0712Z'])),
])
def test__make_aggregate_index_offset_left(length, label, expected):
    test_data = {
        0: pd.DataFrame(range(6), index=pd.date_range(
            '20191004T0658Z', freq='7min', periods=6))  # end 32
        }
    out = utils._make_aggregate_index(test_data, length, label, 'UTC')
    pdt.assert_index_equal(out, expected)


def test__make_aggregate_index_tz():
    length = '30min'
    label = 'beginning'
    test_data = {
        0: pd.DataFrame(range(6), index=pd.date_range(
            '20190101T1600Z', freq='5min', periods=6))  # end 30
        }
    expected = pd.DatetimeIndex(['20190101T0900'],
                                tz='America/Denver')
    out = utils._make_aggregate_index(test_data, length, label,
                                      'America/Denver')
    pdt.assert_index_equal(out, expected)


def test__make_aggregate_index_invalid_length():
    length = '33min'
    label = 'beginning'
    test_data = {
        0: pd.DataFrame(range(6), index=pd.date_range(
            '20190101T0158Z', freq='7min', periods=6))  # end 32
        }
    with pytest.raises(ValueError):
        utils._make_aggregate_index(test_data, length, label, 'UTC')


def test__make_aggregate_index_instant():
    length = '30min'
    label = 'instant'
    test_data = {
        0: pd.DataFrame(range(6), index=pd.date_range(
            '20190101T0100Z', freq='10min', periods=6))  # end 32
        }
    with pytest.raises(ValueError):
        utils._make_aggregate_index(test_data, length, label, 'UTC')


@pytest.mark.parametrize('start,end', [
    (pd.Timestamp('20190101T0000Z'), pd.Timestamp('20190102T0000')),
    (pd.Timestamp('20190101T0000'), pd.Timestamp('20190102T0000Z')),
    (pd.Timestamp('20190101T0000'), pd.Timestamp('20190102T0000')),
])
def test__make_aggregate_index_localization(start, end):
    length = '30min'
    label = 'ending'
    test_data = {
        0: pd.DataFrame(range(1), index=pd.DatetimeIndex([start])),
        1: pd.DataFrame(range(1), index=pd.DatetimeIndex([end])),
        }
    with pytest.raises(TypeError):
        utils._make_aggregate_index(test_data, length, label, 'UTC')


@pytest.mark.parametrize('inp,oup', [
    (pd.DataFrame(dtype=float), pd.Series(dtype=float)),
    (pd.DataFrame(index=pd.DatetimeIndex([]), dtype=float),
     pd.DataFrame(dtype=float)),
    (pd.Series([0, 1]), pd.Series([0, 1])),
    (pd.DataFrame([[0, 1], [1, 2]]), pd.DataFrame([[0, 1], [1, 2]])),
    pytest.param(
        pd.Series([0, 1]),
        pd.Series([0, 1], index=pd.date_range(start='now', freq='1min',
                                              periods=2)),
        marks=pytest.mark.xfail(type=AssertionError, strict=True)),
    pytest.param(
        pd.Series([0, 1]),
        pd.Series([1, 0]),
        marks=pytest.mark.xfail(type=AssertionError, strict=True))
])
def test_sha256_pandas_object_hash(inp, oup):
    assert utils.sha256_pandas_object_hash(inp) == utils.sha256_pandas_object_hash(oup)  # NOQA


def test_listhandler():
    logger = logging.getLogger('testlisthandler')
    handler = utils.ListHandler()
    logger.addHandler(handler)
    logger.setLevel('DEBUG')
    logger.warning('Test it')
    logger.debug('What?')
    out = handler.export_records()
    assert len(out) == 1
    assert out[0].message == 'Test it'
    assert len(handler.export_records(logging.DEBUG)) == 2


def test_listhandler_recreate():
    logger = logging.getLogger('testlisthandler')
    handler = utils.ListHandler()
    logger.addHandler(handler)
    logger.setLevel('DEBUG')
    logger.warning('Test it')
    logger.debug('What?')
    out = handler.export_records()
    assert len(out) == 1
    assert out[0].message == 'Test it'
    assert len(handler.export_records(logging.DEBUG)) == 2

    l2 = logging.getLogger('testlist2')
    h2 = utils.ListHandler()
    l2.addHandler(h2)
    l2.error('Second fail')
    out = h2.export_records()
    assert len(out) == 1
    assert out[0].message == 'Second fail'


def test_hijack_loggers(mocker):
    old_handler = mocker.MagicMock()
    new_handler = mocker.MagicMock()
    mocker.patch('solarforecastarbiter.utils.ListHandler',
                 return_value=new_handler)
    logger = logging.getLogger('testhijack')
    logger.addHandler(old_handler)
    assert logger.handlers[0] == old_handler
    with utils.hijack_loggers(['testhijack']):
        assert logger.handlers[0] == new_handler
    assert logger.handlers[0] == old_handler


@pytest.mark.parametrize('data,freq,expected', [
    (pd.Series(index=pd.DatetimeIndex([]), dtype=float), '5min',
     [pd.Series(index=pd.DatetimeIndex([]), dtype=float)]),
    (pd.Series([1.0], index=pd.DatetimeIndex(['2020-01-01T00:00Z'])),
     '5min',
     [pd.Series([1.0], index=pd.DatetimeIndex(['2020-01-01T00:00Z']))]),
    (pd.Series(
        [1.0, 2.0, 3.0],
        index=pd.date_range('2020-01-01T00:00Z', freq='1h', periods=3)),
     '1h',
     [pd.Series(
         [1.0, 2.0, 3.0],
         index=pd.date_range('2020-01-01T00:00Z', freq='1h', periods=3))]),
    (pd.Series(
        [1.0, 2.0, 4.0],
        index=pd.DatetimeIndex(['2020-01-01T01:00Z', '2020-01-01T02:00Z',
                                '2020-01-01T04:00Z'])),
     '1h',
     [pd.Series(
         [1.0, 2.0],
         index=pd.DatetimeIndex(['2020-01-01T01:00Z', '2020-01-01T02:00Z'])),
      pd.Series(
          [4.0],
          index=pd.DatetimeIndex(['2020-01-01T04:00Z'])),
      ]),
    (pd.Series(
        [1.0, 3.0, 5.0],
        index=pd.DatetimeIndex(['2020-01-01T01:00Z', '2020-01-01T03:00Z',
                                '2020-01-01T05:00Z'])),
     '1h',
     [pd.Series(
         [1.0],
         index=pd.DatetimeIndex(['2020-01-01T01:00Z'])),
      pd.Series(
         [3.0],
         index=pd.DatetimeIndex(['2020-01-01T03:00Z'])),
      pd.Series(
          [5.0],
          index=pd.DatetimeIndex(['2020-01-01T05:00Z'])),
      ]),
    (pd.DataFrame(index=pd.DatetimeIndex([]), dtype=float), '1h',
     [pd.DataFrame(index=pd.DatetimeIndex([]), dtype=float)]),
    (pd.DataFrame(
        {'a': [1.0, 2.0, 4.0], 'b': [11.0, 12.0, 14.0]},
        index=pd.DatetimeIndex(['2020-01-01T01:00Z', '2020-01-01T02:00Z',
                                '2020-01-01T04:00Z'])),
     '1h',
     [pd.DataFrame(
         {'a': [1.0, 2.0], 'b': [11.0, 12.0]},
         index=pd.DatetimeIndex(['2020-01-01T01:00Z', '2020-01-01T02:00Z'])),
      pd.DataFrame(
          {'a': [4.0], 'b': [14.0]},
          index=pd.DatetimeIndex(['2020-01-01T04:00Z'])),
      ]),
    (pd.DataFrame(
        {'_cid': [1.0, 2.0, 4.0], '_cid0': [11.0, 12.0, 14.0]},
        index=pd.DatetimeIndex(['2020-01-01T01:00Z', '2020-01-01T02:00Z',
                                '2020-01-01T04:00Z'])),
     '1h',
     [pd.DataFrame(
         {'_cid': [1.0, 2.0], '_cid0': [11.0, 12.0]},
         index=pd.DatetimeIndex(['2020-01-01T01:00Z', '2020-01-01T02:00Z'])),
      pd.DataFrame(
          {'_cid': [4.0], '_cid0': [14.0]},
          index=pd.DatetimeIndex(['2020-01-01T04:00Z'])),
      ]),
    (pd.DataFrame(
        [[0.0, 1.0], [2.0, 3.0]],
        columns=pd.MultiIndex.from_product([[0], ['a', 'b']]),
        index=pd.DatetimeIndex(['2020-01-01T00:00Z', '2020-01-02T00:00Z'])),
     '12h',
     [pd.DataFrame(
         [[0.0, 1.0]],
         columns=pd.MultiIndex.from_product([[0], ['a', 'b']]),
         index=pd.DatetimeIndex(['2020-01-01T00:00Z'])),
      pd.DataFrame(
          [[2.0, 3.0]],
          columns=pd.MultiIndex.from_product([[0], ['a', 'b']]),
          index=pd.DatetimeIndex(['2020-01-02T00:00Z']))]),
])
def test_generate_continuous_chunks(data, freq, expected):
    lg = list(utils.generate_continuous_chunks(data, freq))
    assert len(lg) == len(expected)
    for i, g in enumerate(lg):
        if isinstance(expected[i], pd.DataFrame):
            pdt.assert_frame_equal(expected[i], g, check_column_type=False)
        else:
            pdt.assert_series_equal(expected[i], g)


@pytest.mark.parametrize('data,freq,err', [
    (pd.Series(dtype=float), '5min', TypeError),
    (pd.DataFrame(dtype=float), '5min', TypeError),
    ([], '5min', TypeError),
    ([], 'no', TypeError),
    (pd.Series(index=pd.DatetimeIndex([]), dtype=float), 'no', ValueError),
    pytest.param(pd.Series(index=pd.DatetimeIndex([]), dtype=float),
                 '5min', TypeError,
                 marks=pytest.mark.xfail(strict=True))
])
def test_generate_continuous_chunks_errs(data, freq, err):
    with pytest.raises(err):
        list(utils.generate_continuous_chunks(data, freq))



"""
Fetch NWP files from NCEP NOMADS for select variables. Should primarily be used
as a CLI program.


The script fetches grib2 files from NOMADS as they are available using g2sub,
uses wgrib2 to convert the grib files to netCDF (and adds wind speed),
and optimizes the netCDF file for our expected usage accessing a time-series
for a single location.

Each possible model has an associated dictonary with parameters that are
passed in the HTTP request. Other keys in the dictionaries are used
to set the timing of the HTTP requests:
- update_freq: the time between full model runs
- valid_hr_gen: A generator to produce the valid forecast hours that the model
  produces
- time_between_fcst_hrs: The time in seconds that it typically takes to produce
  the grib2 file for the next forecasted hour
- delay_to_first_forecast: The time after the nominal initialization time
  that the first forecast from the model is typically available
- avg_max_run_length: The average time it takes the model to produce the last
  possible forecast file from the first forecast file.

Many of these parameters are inferred from
https://www.nco.ncep.noaa.gov/pmb/nwprod/prodstat/


This script uses features of asyncio that are likely not available in Windows.
"""
import asyncio
from itertools import chain
import logging
import os
from pathlib import Path
import re
import shutil
import stat
import subprocess
import sys
import tempfile


import aiohttp
import pandas as pd
import xarray as xr


from solarforecastarbiter.io.fetch import (
    make_session, run_in_executor, abort_all_on_exception)


logger = logging.getLogger(__name__)


CHECK_URL = 'https://nomads.ncep.noaa.gov/pub/data/nccf/com/{}/prod'
BASE_URL = 'https://nomads.ncep.noaa.gov/cgi-bin/'
DOMAIN = {'subregion': '',
          'leftlon': -126,
          'rightlon': -66,
          'toplat': 50,
          'bottomlat': 24}


GFS_0P25_1HR = {'endpoint': 'filter_gfs_0p25_1hr.pl',
                'file': 'gfs.t{init_hr:02d}z.pgrb2.0p25.f{valid_hr:03d}',
                'dir': '/gfs.{init_date}/{init_hr}',
                'lev_2_m_above_ground': 'on',
                'lev_10_m_above_ground': 'on',
                'lev_entire_atmosphere': 'on',
                'lev_surface': 'on',
                'var_DSWRF': 'on',
                'var_TCDC': 'on',
                'var_TMP': 'on',
                'var_UGRD': 'on',
                'var_VGRD': 'on',
                'update_freq': '6h',
                'valid_hr_gen': lambda x: chain(range(120), range(120, 240, 3),
                                                range(240, 385, 12)),
                'time_between_fcst_hrs': 60,
                'delay_to_first_forecast': '200min',
                'avg_max_run_length': '100min',
                'filename': 'gfs_0p25.nc'}


NAM_CONUS = {'endpoint': 'filter_nam.pl',
             'file': 'nam.t{init_hr:02d}z.awphys{valid_hr:02d}.tm00.grib2',
             'dir': '/nam.{init_date}',
             'lev_2_m_above_ground': 'on',
             'lev_10_m_above_ground': 'on',
             'lev_entire_atmosphere_\\(considered_as_a_single_layer\\)': 'on',
             'lev_surface': 'on',
             'var_DSWRF': 'on',
             'var_TCDC': 'on',
             'var_TMP': 'on',
             'var_UGRD': 'on',
             'var_VGRD': 'on',
             'update_freq': '6h',
             'valid_hr_gen': lambda x: chain(range(36), range(36, 85, 3)),
             'time_between_fcst_hrs': 60,
             'delay_to_first_forecast': '90min',
             'avg_max_run_length': '80min',
             'filename': 'nam_12km.nc'}


# should be able to use RANGE requests and get data directly from grib files
# like https://www.cpc.ncep.noaa.gov/products/wesley/fast_downloading_grib.html
# so we can get DSWRF for RAP
RAP = {'endpoint': 'filter_rap.pl',
       'file': 'rap.t{init_hr:02d}z.awp130pgrbf{valid_hr:02d}.grib2',
       'dir': '/rap.{init_date}',
       'lev_2_m_above_ground': 'on',
       'lev_10_m_above_ground': 'on',
       'lev_entire_atmosphere': 'on',
       'lev_surface': 'on',
       'var_TCDC': 'on',
       'var_TMP': 'on',
       'var_UGRD': 'on',
       'var_VGRD': 'on',
       'update_freq': '1h',
       'valid_hr_gen': (
           lambda x: range(40) if x in (3, 9, 15, 21) else range(22)),
       'time_between_fcst_hrs': 60,
       'delay_to_first_forecast': '50min',
       'avg_max_run_length': '30min',
       'filename': 'rap.nc'}


HRRR_HOURLY = {
    'endpoint': 'filter_hrrr_2d.pl',
    'file': 'hrrr.t{init_hr:02d}z.wrfsfcf{valid_hr:02d}.grib2',
    'dir': '/hrrr.{init_date}/conus',
    'lev_2_m_above_ground': 'on',
    'lev_10_m_above_ground': 'on',
    'lev_entire_atmosphere': 'on',
    'lev_surface': 'on',
    'var_DSWRF': 'on',
    'var_VBDSF': 'on',
    'var_VDDSF': 'on',
    'var_TCDC': 'on',
    'var_TMP': 'on',
    'var_UGRD': 'on',
    'var_VGRD': 'on',
    'update_freq': '1h',
    'valid_hr_gen': (
        lambda x: range(37) if x in (0, 6, 12, 18) else range(19)),
    'time_between_fcst_hrs': 120,
    'delay_to_first_forecast': '45min',
    'avg_max_run_length': '70min',
    'filename': 'hrrr_hourly.nc'}


HRRR_SUBHOURLY = {
    'endpoint': 'filter_hrrr_sub.pl',
    'file': 'hrrr.t{init_hr:02d}z.wrfsubhf{valid_hr:02d}.grib2',
    'dir': '/hrrr.{init_date}/conus',
    'lev_2_m_above_ground': 'on',
    'lev_10_m_above_ground': 'on',
    'lev_entire_atmosphere': 'on',
    'lev_surface': 'on',
    'var_DSWRF': 'on',
    'var_VBDSF': 'on',
    'var_VDDSF': 'on',
    'var_TMP': 'on',
    'var_WIND': 'on',
    'update_freq': '1h',
    'valid_hr_gen': (lambda x: range(19)),
    'time_between_fcst_hrs': 120,
    'delay_to_first_forecast': '45min',
    'avg_max_run_length': '50min',
    'filename': 'hrrr_subhourly.nc'}


# each GEFS stat_or_member is treated separately
# really makes use of async capabilities
GEFS_0P50_RAW = {'endpoint': 'filter_gens_0p50.pl',
                 'file': 'ge{stat_or_member}.t{init_hr:02d}z.pgrb2a.0p50.f{valid_hr:03d}',  # NOQA
                 'dir': '/gefs.{init_date}/{init_hr}/pgrb2ap5',
                 'lev_2_m_above_ground': 'on',
                 'lev_10_m_above_ground': 'on',
                 'lev_entire_atmosphere': 'on',
                 'lev_surface': 'on',
                 'var_DSWRF': 'on',
                 'var_TCDC': 'on',
                 'var_TMP': 'on',
                 'var_UGRD': 'on',
                 'var_VGRD': 'on',
                 'update_freq': '6h',
                 'valid_hr_gen': lambda x: chain(range(0, 192, 3),
                                                 range(192, 385, 6)),
                 'time_between_fcst_hrs': 60,
                 'delay_to_first_forecast': '280min',
                 'avg_max_run_length': '60min',
                 'filename': 'gefs_{stat_or_member}.nc',
                 'members': (['avg', 'c00', 'spr'] +
                             [f'p{r:02d}' for r in range(1, 21)]),
                 'check_url_name': 'gens'}


EXTRA_KEYS = ['update_freq', 'valid_hr_gen', 'time_between_fcst_hrs',
              'delay_to_first_forecast', 'avg_max_run_length', 'filename',
              'check_url_name', 'member']

model_map = {'gfs_0p25': GFS_0P25_1HR, 'nam_12km': NAM_CONUS,
             'rap': RAP, 'hrrr_hourly': HRRR_HOURLY,
             'hrrr_subhourly': HRRR_SUBHOURLY,
             'gefs': GEFS_0P50_RAW}

NC_TBL = """
TMP:surface:ignore
TMP:2 m above ground:t2m
UGRD:10 m above ground:ignore
VGRD:10 m above ground:ignore
TCDC:entire atmosphere:tcdc
TCDC:entire atmosphere (considered as a single layer):tcdc
DSWRF:surface:dswrf
VBDSF:surface:vbdsf
VDDSF:surface:vddsf
WIND:10 m above ground:si10
"""

COMPRESSION = {'zlib': True, 'complevel': 1, 'shuffle': True,
               'fletcher32': True}
DEFAULT_ENCODING = {
    # stores the time steps, not an actual time
    'time': {'dtype': 'int16'},
    'latitude': {'dtype': 'float32', 'least_significant_digit': 3},
    'longitude': {'dtype': 'float32', 'least_significant_digit': 3}
}
LEAST_SIGNIFICANT_DIGITS = {
    't2m': 2,
    'tcdc': 1,
    'si10': 2,
    'dswrf': 1,
    'vbdsf': 1,
    'vddsf': 1
}


async def get_with_retries(get_func, *args, retries=5, **kwargs):
    """
    Call get_func and retry if the request fails

    Parameters
    ----------
    get_func : function
        Function that performs an aiohttp call to be retried
    retries : int
        Number of retries before raising the error
    *args
        Passed to get_func
    **kwargs
        Passed to get_func

    Returns
    -------
    Result of get_func

    Raises
    ------
    aiohttp.ClientResponseError
        When get_func fails after retrying retries times
    """
    retried = 0
    while True:
        try:
            res = await get_func(*args, **kwargs)
        except aiohttp.ClientResponseError as e:
            logger.warning('Request to %s failed with code %s, retrying',
                           e.request_info.url, e.status)
            retried += 1
            if retried >= retries:
                raise
        except aiohttp.ClientError:
            logger.warning('Request failed in connection, retrying')
            retried += 1
            if retried >= retries:
                raise
        else:
            return res

        await asyncio.sleep(60)


def _simple_model(model):
    return model['dir'].split('.')[0][1:]


@abort_all_on_exception
async def get_available_dirs(session, model):
    """Get the available date/date+init_hr directories"""
    simple_model = _simple_model(model)
    is_init_date = 'init_date' in model['dir']
    model_url = BASE_URL + model['endpoint']

    async def _get(model_url):
        async with session.get(model_url, raise_for_status=True) as r:
            return await r.text()

    page = await get_with_retries(_get, model_url)
    if is_init_date:
        list_avail_days = set(
            re.findall(simple_model + '\\.([0-9]{8})', page))
    else:
        list_avail_days = set(
            re.findall(simple_model + '\\.([0-9]{10})', page))
    return list_avail_days


def _process_params(model, init_time):
    """Generator to get the parameters for fetching forecasts for a given
    model at a given init_time"""
    params = model.copy()
    params.update(DOMAIN)
    valid_hr_gen = params['valid_hr_gen'](init_time.hour)
    for p in EXTRA_KEYS:
        if p in params:
            del params[p]
    params['dir'] = params['dir'].format(
        init_date=init_time.strftime('%Y%m%d'),
        init_dt=init_time.strftime('%Y%m%d%H'),
        init_hr=init_time.strftime('%H'))
    for i in valid_hr_gen:
        newp = params.copy()
        newp['file'] = newp['file'].format(
            init_hr=init_time.hour,
            valid_hr=i)
        yield newp


async def check_next_inittime(session, init_time, model):
    """Check if data from the next model initializtion time is available"""
    next_inittime = init_time + pd.Timedelta(model['update_freq'])
    simple_model = _simple_model(model)
    next_init_url = (CHECK_URL.format(model.get('check_url_name',
                                                simple_model))
                     + model['dir'].format(
                         init_date=next_inittime.strftime('%Y%m%d'),
                         init_dt=next_inittime.strftime('%Y%m%d%H'),
                         init_hr=next_inittime.strftime('%H'))
                     + '/' + model['file'].format(init_hr=next_inittime.hour,
                                                  valid_hr=0))

    try:
        async with session.head(next_init_url) as r:
            if r.status == 200:
                logger.warning(
                    'Skipping to next init time at %s for %s %s',
                    next_inittime, simple_model, model.get('member', ''))
                return True
            else:
                return False
    except aiohttp.ClientOSError:
        return False


def get_filename(basepath, init_time, params):
    filename = (
        basepath / init_time.strftime('%Y/%m/%d/%H') / params['file'])
    if not filename.suffix == '.grib2':
        filename = filename.with_suffix(filename.suffix + '.grib2')
    return filename


async def files_to_retrieve(session, model, modelpath, init_time):
    """Generator to return the parameters of the available files for download
    """
    possible_params = _process_params(model, init_time)
    simple_model = _simple_model(model)
    first_file_modified_at = None
    for next_params in possible_params:
        logger.debug('Checking if file is available for %s', next_params)
        filename = get_filename(modelpath, init_time, next_params)
        if filename.exists():
            yield next_params
            continue
        next_model_url = (CHECK_URL.format(model.get('check_url_name',
                                                     simple_model))
                          + next_params['dir'] + '/' + next_params['file'])
        while True:
            # is the next file ready?
            try:
                logger.debug('Calling HEAD %s', next_model_url)
                async with session.head(
                        next_model_url, raise_for_status=True) as r:
                    if first_file_modified_at is None:
                        first_file_modified_at = pd.Timestamp(
                            r.headers['Last-Modified'])
                        logger.debug('First file was available at %s %s',
                                     first_file_modified_at,
                                     model.get('member', ''))
                logger.debug('HEAD returned %s', next_model_url)
            except aiohttp.ClientResponseError as e:
                if e.status == 404:  # Not found
                    logger.debug(
                        'Next file not ready yet for %s at %s %s\n%s %s',
                        simple_model, init_time, model.get('member', ''),
                        e.status, e.message)
                else:
                    logger.error(
                        'Error checking if next file is ready %s\n'
                        '%s %s', model.get('member', ''), e.status, e.message)
            except aiohttp.ClientError as e:
                logger.warning('Error in checking for next file %s %s',
                               model.get('member', ''), str(e))
            else:
                logger.debug('%s/%s is ready for download',
                             next_params['dir'], next_params['file'])
                yield next_params
                break

            # if the current time is after 'avg_max_run_length' after the
            # first forecast was available, check if forecasts from the
            # next model run are available and if so, move on to that run
            if (
                    first_file_modified_at is not None and
                    pd.Timestamp.utcnow() > first_file_modified_at +
                    pd.Timedelta(model['avg_max_run_length'])
            ):
                nextrun_available = await check_next_inittime(
                    session, init_time, model)
                if nextrun_available:
                    return
            await asyncio.sleep(model['time_between_fcst_hrs'])


async def _get_file(session, url, params, tmpfile, chunksize):
    timeout = aiohttp.ClientTimeout(total=660, connect=60, sock_read=600)
    async with session.get(url, params=params, raise_for_status=True,
                           timeout=timeout) as r:
        with open(tmpfile, 'wb') as f:
            async for chunk in r.content.iter_chunked(chunksize * 1024):
                f.write(chunk)


@abort_all_on_exception
async def fetch_grib_files(session, params, basepath, init_time, chunksize):
    """
    Fetch the grib file referenced by params and save to the appropriate
    folder under basepath. Retrieves the files in chunks.

    Parameters
    ----------
    session : aiohttp.ClientSession
        The HTTP session to use to request the file
    params : dict
        Parameters to include in the GET query to params['endpoint']
    basepath : Path
        Path to the base directory where files will be saved. New directories
        under basepath of the form basepath / year / month / day / hour
        will be created as necessary.
    init_time : datetime
        Initialization time of the model we're trying to fetch
    chunksize : int
        Chunksize in KB to fetch and save at once

    Returns
    -------
    filename : Path
        Path of the successfully saved file

    Raises
    ------
    aiohttp.ClientResponseError
        When the HTTP request fails/returns a status code >= 400
    """
    endpoint = params.pop('endpoint')
    url = BASE_URL + endpoint
    filename = get_filename(basepath, init_time, params)
    if filename.exists():
        return filename
    if not filename.parent.is_dir():
        filename.parent.mkdir(parents=True)
    logger.info('Getting file %s', filename)
    tmpfile = filename.with_name('.tmp_' + filename.name)
    await get_with_retries(_get_file, session, url, params, tmpfile, chunksize)
    tmpfile.rename(filename)
    logging.debug('Successfully saved %s', filename)
    return filename


@abort_all_on_exception
async def process_grib_to_netcdf(folder, model):
    logger.info('Converting GRIB files to NetCDF with wgrib2 %s',
                model.get('member', ''))
    _handle, nctmp = tempfile.mkstemp()
    os.close(_handle)
    nctmp = Path(nctmp)
    # possible that this holds up processing on file io
    # so run in separate process
    grib_prefix = model['file'].split('.')[0]
    wind_in_model = 'var_WIND' not in model
    try:
        await run_in_executor(_process_grib, folder, nctmp, grib_prefix,
                              wind_in_model)
    except Exception:
        nctmp.unlink()
        raise
    return nctmp


def _process_grib(folder, nctmp, grib_prefix, wind_in_model):
    if wind_in_model:
        # need to add wind to the grib files
        for grbfile in folder.glob(f'{grib_prefix}*.grib2'):
            path = str(grbfile.resolve())
            try:
                subprocess.run(
                    f'wgrib2 {path} -wind_speed - -match "(UGRD|VGRD)" | '
                    f'wgrib2 - -append -grib_out {path}',
                    shell=True, check=True, capture_output=True)
            except subprocess.CalledProcessError as e:
                logger.error('Error converting wind in file %s\n%s',
                             grbfile, e.stderr)
                raise OSError

    if 'subhourly' in str(folder):
        # for hrrr subhourly, assume TMP and VDDSF have no average but others
        fmt = "-match 'ave|TMP|VDDSF'"
    else:
        fmt = ''

    with tempfile.NamedTemporaryFile(mode='w') as tmp_nc_tbl:
        tmp_nc_tbl.write(NC_TBL)
        tmp_nc_tbl.flush()

        try:
            subprocess.run(
                f'cat {str(folder)}/{grib_prefix}*.grib2 | '
                f'wgrib2 - -nc4 -nc_table {tmp_nc_tbl.name} {fmt} -append -netcdf {str(nctmp)}',  # NOQA
                shell=True, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            logger.error('Error converting grib files %s*.grib2 to netCDF\n%s',
                         grib_prefix, e.stderr)
            raise OSError
    return nctmp


def _optimize_netcdf(nctmpfile, out_path):
    """Optmizes the netcdf file for accessing by time slice."""
    ds = xr.open_dataset(nctmpfile, engine='netcdf4',
                         backend_kwargs={'mode': 'r'})
    # time is likely unlimited
    if 'unlimited_dims' in ds.encoding:
        ds.encoding = {}

    chunksizes = []
    for dim, size in ds.dims.items():
        if dim == 'time':
            chunksizes.append(size)
        else:
            chunksizes.append(50)

    encoding = DEFAULT_ENCODING.copy()
    encoding.update(
        {key: {'dtype': 'float32',
               'least_significant_digit': LEAST_SIGNIFICANT_DIGITS[key],
               'chunksizes': chunksizes,
               **COMPRESSION}
         for key in ds.keys()})
    ds.to_netcdf(out_path, format='NETCDF4',
                 mode='w', unlimited_dims=None,
                 encoding=encoding)
    ds.close()


async def optimize_netcdf(nctmpfile, final_path):
    """Compress the netcdf file and adjust the chunking for fast time-series
    access"""
    logger.info('Optimizing NetCDF file to save at %s', final_path)
    parent = Path(final_path.parent)
    if not parent.is_dir():
        parent.mkdir(parents=True)
    _handle, tmp_path = tempfile.mkstemp(dir=parent)
    os.close(_handle)
    tmp_path = Path(tmp_path)
    # possible that this leaks memory, so run in separate process
    # that is restarted after a number of jobs
    try:
        await run_in_executor(_optimize_netcdf, nctmpfile, tmp_path)
    except Exception:
        tmp_path.unlink()
        raise
    else:
        tmp_path.rename(final_path)
        final_path.chmod(stat.S_IRGRP | stat.S_IRUSR | stat.S_IROTH |
                         stat.S_IWUSR)
        logger.info('Done optimizing NetCDF at %s', final_path)
    finally:
        nctmpfile.unlink()


async def sleep_until_inittime(inittime, model):
    # don't bother requesting a file until it might be ready
    now = pd.Timestamp.utcnow()
    likely_ready_time = inittime + pd.Timedelta(
        model['delay_to_first_forecast'])
    if likely_ready_time > now:
        seconds = (likely_ready_time - now).total_seconds()
        logger.info('Sleeping %0.1fs for next model run %s', seconds,
                    model.get('member', ''))
        await asyncio.sleep(seconds)


async def startup_find_next_runtime(model_path, session, model):
    """Find the next model run to get based on what is available
    on NOMADS and what .nc files are present locally"""
    dirs = await get_available_dirs(session, model)
    no_file = []
    first = pd.Timestamp('20000101T0000Z')
    max_time = first
    for dir_ in dirs:
        if len(dir_) == 8:
            path = model_path / dir_[:4] / dir_[4:6] / dir_[6:8]
            for hr in range(0, 24, int(model['update_freq'].strip('h'))):
                hrpath = path / f'{hr:02d}'
                hrtime = pd.Timestamp(f'{dir_[:8]}T{hr:02d}00Z')
                if not (hrpath / model['filename']).exists():
                    no_file.append(hrtime)
                else:
                    max_time = max(max_time, hrtime)
        else:
            hrpath = model_path / dir_[:4] / dir_[4:6] / dir_[6:8] / dir_[8:10]
            hrtime = pd.Timestamp(f'{dir_[:8]}T{dir_[8:10]}00Z')
            if not (hrpath / model['filename']).exists():
                no_file.append(hrtime)
            else:
                max_time = max(max_time, hrtime)
    if len(no_file) == 0:
        if max_time > first:
            inittime = max_time + pd.Timedelta(model['update_freq'])
        else:  # No available dirs?
            raise ValueError(
                'Failed to find next available model from NOMADS %s' %
                model.get('member', ''))
    else:
        inittime = min(no_file)
    await sleep_until_inittime(inittime, model)
    return inittime


async def next_run_time(inittime, modelpath, model):
    inittime += pd.Timedelta(model['update_freq'])
    # check if nc file exists for this inittime
    if (
            modelpath / inittime.strftime('%Y/%m/%d/%H') /
            model['filename']).exists():
        return await next_run_time(inittime, modelpath, model)
    await sleep_until_inittime(inittime, model)
    return inittime


async def _run_loop(session, model, modelpath, chunksize, once, use_tmp):
    inittime = await startup_find_next_runtime(modelpath, session, model)
    while True:
        fetch_tasks = set()
        finalpath = (modelpath / inittime.strftime('%Y/%m/%d/%H') /
                     model['filename'])
        if use_tmp:
            _tmpdir = tempfile.TemporaryDirectory()
            gribdir = Path(_tmpdir.name)
        else:
            gribdir = modelpath
        async for params in files_to_retrieve(session, model, gribdir,
                                              inittime):
            logger.debug('Processing parameters %s', params)
            fetch_tasks.add(asyncio.create_task(
                fetch_grib_files(session, params, gribdir, inittime,
                                 chunksize)))
        files = await asyncio.gather(*fetch_tasks)
        if len(files) != 0:  # skip to next inittime
            path_to_files = files[0].parent
            try:
                nctmpfile = await process_grib_to_netcdf(path_to_files,
                                                         model)
                await optimize_netcdf(nctmpfile, finalpath)
            except Exception:
                raise
        if use_tmp:
            _tmpdir.cleanup()
        else:
            # remove grib files
            for f in files:
                f.unlink()
        if once:
            break
        else:
            logger.info('Moving on to next model run')
            inittime = await next_run_time(inittime, modelpath, model)


async def run(basepath, model_name, chunksize, once=False, use_tmp=False):
    session = make_session()
    modelpath = basepath / model_name
    if model_name != 'gefs':
        model = model_map[model_name]
        await _run_loop(session, model, modelpath, chunksize, once, use_tmp)
    else:
        base_model = model_map[model_name].copy()
        members = base_model.pop('members')
        member_loops = set()
        for member in members:
            model = base_model.copy()
            model['member'] = member
            model['file'] = model['file'].replace('{stat_or_member}', member)
            model['filename'] = model['filename'].format(stat_or_member=member)
            member_loops.add(asyncio.create_task(
                _run_loop(session, model, modelpath, chunksize, once,
                          use_tmp)))
        await asyncio.wait(member_loops)
    await session.close()


async def optimize_only(path_to_files, model_name):
    model = model_map[model_name]
    nctmpfile = await process_grib_to_netcdf(path_to_files, model)
    try:
        await optimize_netcdf(
            nctmpfile, path_to_files / f'{model_name}.nc')
    except Exception:
        raise
    else:
        # remove grib files
        for f in path_to_files.glob(f'{model["file"].split(".")[0]}*.grib2'):
            f.unlink()


def check_wgrib2():
    if shutil.which('wgrib2') is None:
        logger.error('wgrib2 was not found in PATH and is required')
        sys.exit(1)



import re
import shutil
import subprocess

import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset
import pytest
import jinja2
from bokeh import __version__ as bokeh_version
from plotly import __version__ as plotly_version

from solarforecastarbiter import datamodel
from solarforecastarbiter.reports import template


expected_metrics_json = """[{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"total","metric":"mae","value":2,"index":1},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"date","metric":"mae","value":2,"index":1546300800000},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"hour","metric":"mae","value":2,"index":1},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"total","metric":"rmse","value":2,"index":1},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"date","metric":"rmse","value":2,"index":1546300800000},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"hour","metric":"rmse","value":2,"index":1},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"total","metric":"mbe","value":2,"index":1},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"date","metric":"mbe","value":2,"index":1546300800000},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"hour","metric":"mbe","value":2,"index":1},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"total","metric":"s","value":2,"index":1},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"date","metric":"s","value":2,"index":1546300800000},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"hour","metric":"s","value":2,"index":1},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"total","metric":"cost","value":2,"index":1},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"date","metric":"cost","value":2,"index":1546300800000},{"name":"0 Day GFS GHI","abbrev":"0 Day GFS GHI","category":"hour","metric":"cost","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"total","metric":"mae","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"date","metric":"mae","value":2,"index":1546300800000},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"hour","metric":"mae","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"total","metric":"rmse","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"date","metric":"rmse","value":2,"index":1546300800000},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"hour","metric":"rmse","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"total","metric":"mbe","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"date","metric":"mbe","value":2,"index":1546300800000},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"hour","metric":"mbe","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"total","metric":"s","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"date","metric":"s","value":2,"index":1546300800000},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"hour","metric":"s","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"total","metric":"cost","value":2,"index":1},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"date","metric":"cost","value":2,"index":1546300800000},{"name":"Day Ahead GFS GHI","abbrev":"Day Ahe. GFS GHI","category":"hour","metric":"cost","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"total","metric":"mae","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"date","metric":"mae","value":2,"index":1546300800000},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"hour","metric":"mae","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"total","metric":"rmse","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"date","metric":"rmse","value":2,"index":1546300800000},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"hour","metric":"rmse","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"total","metric":"mbe","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"date","metric":"mbe","value":2,"index":1546300800000},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"hour","metric":"mbe","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"total","metric":"s","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"date","metric":"s","value":2,"index":1546300800000},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"hour","metric":"s","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"total","metric":"cost","value":2,"index":1},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"date","metric":"cost","value":2,"index":1546300800000},{"name":"GHI Aggregate FX 60","abbrev":"GHI Agg. FX 60","category":"hour","metric":"cost","value":2,"index":1}]"""  # NOQA


@pytest.fixture
def mocked_timeseries_plots(mocker):
    mocked = mocker.patch(
        'solarforecastarbiter.reports.figures.plotly_figures.timeseries_plots')
    mocked.return_value = ('{}', '{}', '{}', False)


@pytest.fixture
def mocked_timeseries_plots_exception(mocker):
    mocked = mocker.patch(
        'solarforecastarbiter.reports.figures.plotly_figures.timeseries_plots')
    mocked.side_effect = Exception


@pytest.fixture
def dash_url():
    return 'https://solarforecastarbiter.url'


@pytest.fixture(params=[True, False])
def with_series(request):
    return request.param


@pytest.fixture
def expected_kwargs(dash_url):
    def fn(report, with_series, with_report=True):
        kwargs = {}
        kwargs['human_categories'] = datamodel.ALLOWED_CATEGORIES
        kwargs['human_metrics'] = datamodel.ALLOWED_METRICS
        kwargs['category_blurbs'] = datamodel.CATEGORY_BLURBS
        if with_report:
            kwargs['report'] = report
        if report.status == 'complete':
            kwargs['metrics_json'] = expected_metrics_json
        else:
            kwargs['metrics_json'] = '[]'
        kwargs['dash_url'] = dash_url
        kwargs['bokeh_version'] = bokeh_version
        kwargs['plotly_version'] = plotly_version
        if with_series:
            kwargs['timeseries_spec'] = '{}'
            kwargs['scatter_spec'] = '{}'
            kwargs['timeseries_prob_spec'] = '{}'
            kwargs['includes_distribution'] = False
        return kwargs
    return fn


def test__get_render_kwargs_no_series(
        mocked_timeseries_plots, report_with_raw, dash_url, with_series,
        expected_kwargs):
    kwargs = template._get_render_kwargs(
        report_with_raw,
        dash_url,
        with_series
    )
    exp = expected_kwargs(report_with_raw, with_series)
    assert kwargs == exp


def test__get_render_kwargs_pending(
        mocked_timeseries_plots, pending_report, dash_url,
        expected_kwargs, mocker):
    kwargs = template._get_render_kwargs(
        pending_report,
        dash_url,
        False
    )
    exp = expected_kwargs(pending_report, False)
    assert kwargs == exp


def test__get_render_kwargs_with_series_exception(
        report_with_raw, dash_url, mocked_timeseries_plots_exception):
    kwargs = template._get_render_kwargs(
        report_with_raw,
        dash_url,
        True
    )
    assert kwargs['timeseries_spec'] == '{}'
    assert kwargs['scatter_spec'] == '{}'
    assert kwargs['timeseries_prob_spec'] == '{}'


@pytest.fixture(params=[0, 1, 2])
def good_or_bad_report(request, report_with_raw, failed_report,
                       pending_report):
    if request.param == 0:
        out = report_with_raw
    elif request.param == 1:
        out = failed_report
    elif request.param == 2:
        out = pending_report
    return out


@pytest.mark.parametrize('with_body', [True, False])
def test_get_template_and_kwargs(
        good_or_bad_report, dash_url, with_series, expected_kwargs,
        mocked_timeseries_plots, with_body):
    html_template, kwargs = template.get_template_and_kwargs(
        good_or_bad_report,
        dash_url,
        with_series,
        with_body
    )
    base = kwargs.pop('base_template')
    kwargs.pop('report')
    assert type(base) == jinja2.environment.Template
    assert type(html_template) == jinja2.environment.Template
    assert kwargs == expected_kwargs(good_or_bad_report,
                                     with_series, False)


def test_get_template_and_kwargs_bad_status(
        report_with_raw, dash_url, mocked_timeseries_plots):
    inp = report_with_raw.replace(status='notokay')
    with pytest.raises(ValueError):
        template.get_template_and_kwargs(
            inp, dash_url, False, True)


def test_render_html_body_only(report_with_raw, dash_url, with_series,
                               mocked_timeseries_plots):
    rendered = template.render_html(
        report_with_raw, dash_url, with_series, True)
    assert rendered[:22] == '<h1 id="report-title">'


def test_render_html_full_html(report_with_raw, dash_url, with_series,
                               mocked_timeseries_plots):
    rendered = template.render_html(
        report_with_raw, dash_url, with_series, False)
    assert rendered[:46] == '<!doctype html>\n<html lang="en" class="h-100">'


def test_build_metrics_json():
    pass


@pytest.mark.parametrize('val,expected', [
    ('<p> paragraph</p>', ' paragraph\n'),
    ('<em>italic</em>', '\\emph{italic}'),
    ('<code>nan</code>', '\\verb|nan|'),
    ('<b>bold</b>', '\\textbf{bold}'),
    ('<ol>\n<li>item one</li>\n</ol>',
     '\\begin{enumerate}\n\\item item one\n\n\\end{enumerate}'),
    ('<a href="tolink" class="what">stuff</a>', 'stuff'),
    (('<p>paragraph one <em>important</em> code here <code>null</code>'
      ' and more <b>bold</b><em> critical</em> <code>here</code></p>'
      ' <b>masbold</b>'),
     ('paragraph one \\emph{important} code here \\verb|null|'
      ' and more \\textbf{bold}\\emph{ critical} \\verb|here|\n'
     ' \\textbf{masbold}'))
])
def test_html_to_tex(val, expected):
    assert template._html_to_tex(val) == expected


def test_render_pdf(report_with_raw, dash_url):
    if shutil.which('pdflatex') is None:  # pragma: no cover
        pytest.skip('pdflatex must be on PATH to generate PDF reports')
    rendered = template.render_pdf(report_with_raw, dash_url)
    assert rendered.startswith(b'%PDF')


def test_render_pdf_special_chars(
        ac_power_observation_metadata, ac_power_forecast_metadata, dash_url,
        fail_pdf, preprocessing_result_types, report_metrics):
    if shutil.which('pdflatex') is None:  # pragma: no cover
        pytest.skip('pdflatex must be on PATH to generate PDF reports')
    quality_flag_filter = datamodel.QualityFlagFilter(
        (
            "USER FLAGGED",
        )
    )
    forecast = ac_power_forecast_metadata.replace(
        name="ac_power forecast (why,)  ()'-_,")
    observation = ac_power_observation_metadata.replace(
        name="ac_power observations  ()'-_,")
    fxobs = datamodel.ForecastObservation(forecast,
                                          observation)
    tz = 'America/Phoenix'
    start = pd.Timestamp('20190401 0000', tz=tz)
    end = pd.Timestamp('20190404 2359', tz=tz)
    report_params = datamodel.ReportParameters(
        name="NREL MIDC OASIS GHI Forecast Analysis  ()'-_,",
        start=start,
        end=end,
        object_pairs=(fxobs,),
        metrics=("mae", "rmse", "mbe", "s"),
        categories=("total", "date", "hour"),
        filters=(quality_flag_filter,)
    )
    report = datamodel.Report(
        report_id="56c67770-9832-11e9-a535-f4939feddd83",
        report_parameters=report_params
    )
    qflags = list(
        f.quality_flags for f in report.report_parameters.filters if
        isinstance(f, datamodel.QualityFlagFilter)
    )
    qflags = list(qflags[0])
    ser_index = pd.date_range(
        start, end,
        freq=to_offset(forecast.interval_length),
        name='timestamp')
    ser = pd.Series(
        np.repeat(100, len(ser_index)), name='value',
        index=ser_index)
    pfxobs = datamodel.ProcessedForecastObservation(
        forecast.name,
        fxobs,
        forecast.interval_value_type,
        forecast.interval_length,
        forecast.interval_label,
        valid_point_count=len(ser),
        validation_results=tuple(datamodel.ValidationResult(
            flag=f, count=0) for f in qflags),
        preprocessing_results=tuple(datamodel.PreprocessingResult(
            name=t, count=0) for t in preprocessing_result_types),
        forecast_values=ser,
        observation_values=ser
    )

    figs = datamodel.RawReportPlots(
        (
            datamodel.PlotlyReportFigure.from_dict(
                {
                    'name': 'mae tucson ac_power',
                    'spec': '{"data":[{"x":[1],"y":[1],"type":"bar"}]}',
                    'pdf': fail_pdf,
                    'figure_type': 'bar',
                    'category': 'total',
                    'metric': 'mae',
                    'figure_class': 'plotly',
                }
            ),), '4.5.3',
    )
    raw = datamodel.RawReport(
        generated_at=report.report_parameters.end,
        timezone=tz,
        plots=figs,
        metrics=report_metrics(report),
        processed_forecasts_observations=(pfxobs,),
        versions=(('test',  'test_with_underscore?'),),
        messages=(datamodel.ReportMessage(
            message="Failed to make metrics for ac_power forecast ()'-_,",
            step='', level='', function=''),))
    rr = report.replace(raw_report=raw)
    rendered = template.render_pdf(rr, dash_url)
    assert rendered.startswith(b'%PDF')


def test_render_pdf_not_settled(report_with_raw, dash_url):
    if shutil.which('pdflatex') is None:  # pragma: no cover
        pytest.skip('pdflatex must be on PATH to generate PDF reports')
    with pytest.raises(RuntimeError):
        template.render_pdf(report_with_raw, dash_url, 1)


def test_render_pdf_process_error(report_with_raw, dash_url, mocker):
    mocker.patch('solarforecastarbiter.reports.template.subprocess.run',
                 side_effect=subprocess.CalledProcessError(
                     cmd='', returncode=1))
    with pytest.raises(subprocess.CalledProcessError):
        template.render_pdf(report_with_raw, dash_url)


@pytest.mark.parametrize('text,exp', [
    ('stuf <a href="http://">link</a>', 'stuf \\href{http://}{link}'),
    ('a bunch\nstuf <a href="https://blah">\nlink\nmore\n</a>',
     'a bunch\nstuf \\href{https://blah}{\nlink\nmore\n}'),
    ('no link at all', 'no link at all')
])
def test_link_filter(text, exp):
    new = template._link_filter(text)
    assert new == exp


NOTWORD = re.compile('[^\\w-]')


def test_not_word():
    assert NOTWORD.match('-') is None
    assert NOTWORD.match('test') is None
    assert NOTWORD.match('+') is not None
    assert NOTWORD.match('*') is not None
    assert NOTWORD.match('^') is not None
    assert NOTWORD.match('%') is not None


@pytest.mark.parametrize('val', [
    'ac_power Prob(f <= 10.0%)',
    'ac_power Prob(f <= 10MW) = 10.99%',
    'ac_power Prob(f >= -all)',
    '*!@#$%?.}{[]}<>,."',
    'testit - now\\ W/m^2'
])
def test_figure_name_filter(val):
    new = template._figure_name_filter(val)
    match = NOTWORD.match(new)
    assert match is None



from functools import partial
from pathlib import Path
import types

import pandas as pd
# from pandas.util.testing import assert_series_equal

import pytest

from solarforecastarbiter.io import nwp
from solarforecastarbiter.reference_forecasts import models


latitude = 32.2
longitude = -110.9
elevation = 700
init_time = pd.Timestamp('20190515T0000Z')
start = pd.Timestamp('20190515T0700Z')
end_short = pd.Timestamp('20190515T1200Z')
# gfs is longer, but mixed intervals fails
end_long = pd.Timestamp('20190520T0000Z')

xfail_g2sub = pytest.mark.xfail(reason='ghi does not exist in g2sub')

BASE_PATH = Path(nwp.__file__).resolve().parents[0] / 'tests/data'
LOAD_FORECAST = partial(nwp.load_forecast, base_path=BASE_PATH)


@pytest.mark.parametrize('model', [
    pytest.param(models.gfs_quarter_deg_3hour_to_hourly_mean,
                 marks=pytest.mark.xfail(reason='gfs_3h not available')),
    pytest.param(models.rap_ghi_to_instantaneous, marks=xfail_g2sub)
])
def test_default_load_forecast_failures(model):
    model(
        latitude, longitude, elevation, init_time, start, end_long,
        load_forecast=LOAD_FORECAST)


def check_out(out, start, end, end_strict=True):
    # check times
    for o in out[0:5]:
        assert isinstance(o, (pd.Series, pd.DataFrame))
        assert o.index[0] == start
        if end_strict:
            assert o.index[-1] == end
        else:
            assert o.index[-1] <= end
    # check irradiance limits
    for o in out[0:3]:
        assert (o >= 0).all().all() and (o < 1300).all().all()
    # check temperature limits
    assert (out[3] > -40).all().all() and (out[3] < 60).all().all()
    # check wind speed limits
    assert (out[4] >= 0).all().all() and (out[4] < 60).all().all()
    # check resampling function
    assert isinstance(out[5], (types.FunctionType, partial))
    assert isinstance(out[6], (types.FunctionType, partial))


@pytest.mark.parametrize('interval_label', ['beginning', 'ending'])
@pytest.mark.parametrize('model', [
    models.gfs_quarter_deg_hourly_to_hourly_mean,
    models.gfs_quarter_deg_to_hourly_mean,
    models.hrrr_subhourly_to_hourly_mean,
    models.nam_12km_cloud_cover_to_hourly_mean,
    models.rap_cloud_cover_to_hourly_mean,
    models.gefs_half_deg_to_hourly_mean
])
@pytest.mark.parametrize('end,end_strict', [
    (end_short, True), (end_long, False)
])
def test_mean_models(model, end, end_strict, interval_label):
    out = model(
        latitude, longitude, elevation, init_time, start, end,
        interval_label, load_forecast=LOAD_FORECAST)
    if interval_label == 'beginning':
        start_fx_expected = start
        end_fx_expected = pd.Timestamp(end) - pd.Timedelta('5min')
    elif interval_label == 'ending':
        start_fx_expected = pd.Timestamp(start) + pd.Timedelta('5min')
        end_fx_expected = end
    check_out(out, start_fx_expected, end_fx_expected, end_strict=end_strict)


@pytest.mark.parametrize('model, end_fx_expected', [
    (models.nam_12km_hourly_to_hourly_instantaneous, '20190515T1100Z'),
    (models.hrrr_subhourly_to_subhourly_instantaneous, '20190515T1145Z')
])
def test_instant_models(model, end_fx_expected):
    out = model(
        latitude, longitude, elevation, init_time, start, end_short,
        'instant', load_forecast=LOAD_FORECAST)
    check_out(out, start, pd.Timestamp(end_fx_expected), end_strict=True)


@pytest.mark.parametrize('latitude', [32.0, 32.25, 32.5])
@pytest.mark.parametrize('longitude', [
    -111.0, -110.75, -110.5, -110.25, -110.0])
@pytest.mark.parametrize('start,end,init_time', [
    ('20190515T0100Z', '20190520T0000Z', '20190515T0000Z'),
    ('20190520T0300Z', '20190522T0000Z', '20190515T0000Z'),
    ('20190525T1200Z', '20190531T0000Z', '20190515T0000Z'),
    ('20190715T0100Z', '20190720T0000Z', '20190715T0000Z'),
    ('20190716T0800Z', '20190717T0700Z', '20190715T0000Z'),
])
def test_gfs_quarter_deg_to_hourly_mean(latitude, longitude, start, end,
                                        init_time):
    # Separate test because date ranges are tricky due to GFS output
    # switching from hourly to 3 hourly to 12 hourly.
    # Also ensures that all data is valid at all times (unmixing cloud
    # cover once caused a problem due to rounding beyond our control).
    start = pd.Timestamp(start)
    end = pd.Timestamp(end)
    init_time = pd.Timestamp(init_time)
    out = models.gfs_quarter_deg_to_hourly_mean(
        latitude, longitude, elevation, init_time, start, end,
        'beginning', load_forecast=LOAD_FORECAST)
    # account for beginning interval label
    end_fx_expected = pd.Timestamp(end) - pd.Timedelta('5min')
    check_out(out, start, end_fx_expected, end_strict=True)


@pytest.mark.parametrize('start,end,init_time', [
    ('20190515T0100Z', '20190520T0000Z', '20190515T0000Z'),
    ('20190520T0300Z', '20190522T0000Z', '20190515T0000Z'),
    ('20190525T1200Z', '20190531T0000Z', '20190515T0000Z'),
    ('20190515T0300Z', '20190531T0000Z', '20190515T0000Z'),
])
def test_gefs_half_deg_to_hourly_mean(start, end, init_time):
    start = pd.Timestamp(start)
    end = pd.Timestamp(end)
    init_time = pd.Timestamp(init_time)
    out = models.gefs_half_deg_to_hourly_mean(
        latitude, longitude, elevation, init_time, start, end,
        'beginning', load_forecast=LOAD_FORECAST)
    end_fx_expected = pd.Timestamp(end) - pd.Timedelta('5min')
    check_out(out, start, end_fx_expected, end_strict=True)


@pytest.mark.parametrize('model', [
    'hrrr_hourly',
    'hrrr_subhourly',
    'gfs_0p25',
    'rap',
    'nam_12km',
    'gefs_c00',
    'gefs_p01'
])
def test_domain_limits(model):
    # test file has longitudes ranging from -110.50 to -110.35
    # at midlatitudes, a degree of longitude is approximately 85 km
    # a 10 degree difference is then 850 km.
    # nwp.load_forecast calls lat/lon look up with max dist of 500 km
    with pytest.raises(ValueError):
        LOAD_FORECAST(latitude, -120.5, init_time, start, end_short, model)


@pytest.mark.parametrize('model,exp', [
    (models.gfs_quarter_deg_hourly_to_hourly_mean, 'gfs_0p25'),
    (models.gfs_quarter_deg_to_hourly_mean, 'gfs_0p25'),
    (models.hrrr_subhourly_to_hourly_mean, 'hrrr_subhourly'),
    (models.hrrr_subhourly_to_subhourly_instantaneous, 'hrrr_subhourly'),
    (models.nam_12km_cloud_cover_to_hourly_mean, 'nam_12km'),
    (models.nam_12km_hourly_to_hourly_instantaneous, 'nam_12km'),
    (models.rap_cloud_cover_to_hourly_mean, 'rap'),
    (models.rap_ghi_to_instantaneous, 'rap'),
    (models.gefs_half_deg_to_hourly_mean, 'gefs')
])
def test_get_nwp_model(model, exp):
    assert models.get_nwp_model(model) == exp


@pytest.mark.parametrize('end,end_ceil', [
    ('00Z', '00Z'), ('01Z', '06Z')
])
def test_adjust_gfs_start_end_end(end, end_ceil):
    start = pd.Timestamp('20190101 00Z')
    end = pd.Timestamp(f'20190101 {end}')
    end_ceil = pd.Timestamp(f'20190101 {end_ceil}')
    _, end_ceil_out = models._adjust_gfs_start_end(start, end)
    assert end_ceil_out == end_ceil


@pytest.mark.parametrize('start,start_floor', [
    ('01Z', '01Z'), ('06Z', '01Z'), ('07Z', '07Z')
])
def test_adjust_gfs_start_end_start(start, start_floor):
    end = pd.Timestamp('20190102 00Z')
    start = pd.Timestamp(f'20190101 {start}')
    start_floor = pd.Timestamp(f'20190101 {start_floor}')
    start_floor_out, _ = models._adjust_gfs_start_end(start, end)
    assert start_floor_out == start_floor



import datetime as dt
import json
import re
import uuid


import numpy as np
import pandas as pd
import pytest
from requests.exceptions import HTTPError


from solarforecastarbiter.datamodel import (
    Site, Observation, Forecast, ProbabilisticForecast)
from solarforecastarbiter.io.reference_observations import common
from solarforecastarbiter.io.reference_observations.tests.conftest import (
    expected_site,
    site_dicts,
    site_string_dicts,
    site_objects)


@pytest.fixture
def log(mocker):
    log = mocker.patch('solarforecastarbiter.io.reference_observations.'
                       'common.logger')
    return log


site_test_observation = Observation.from_dict({
    'name': 'site ghi',
    'variable': 'ghi',
    'interval_label': 'ending',
    'interval_value_type': 'interval_mean',
    'interval_length': 1,
    'site': site_objects[0],
    'uncertainty': 0,
    'extra_parameters': ('{"network": "DOE ARM", '
                         '"observation_interval_length": 1,'
                         '"network_api_id": "api_id", '
                         '"network_api_abbreviation": "siteghi"}')
})
invalid_params = {
    'name': 'site-invalid-jsonparams',
    'latitude': 3,
    'longitude': -3,
    'elevation': 6,
    'timezone': 'Etc/GMT+8',
    'extra_parameters': '{{ mon$stertruck',
}
no_params = {
    'name': 'site-invalid-jsonparams',
    'latitude': 3,
    'longitude': -3,
    'elevation': 6,
    'timezone': 'Etc/GMT+8',
}


def test_decode_extra_parameters():
    metadata = Site.from_dict(site_string_dicts[0])
    params = common.decode_extra_parameters(metadata)
    assert params['network'] == 'DOE ARM'
    assert params['observation_interval_length'] == 1


@pytest.mark.parametrize('site', [
    (invalid_params),
    (no_params),
])
def test_decode_extra_parameters_error(site):
    with pytest.raises(ValueError):
        common.decode_extra_parameters(Site.from_dict(site))


@pytest.mark.parametrize('site,expected', [
    (site_objects[0], 'site'),
    (site_objects[1], 'site2'),
    (site_objects[2], 'site3'),
    (site_objects[3], 'site4'),
])
def test_site_name_no_network(site, expected):
    assert common.site_name_no_network(site) == expected


@pytest.mark.parametrize('name,expected', [
    ('{n}_(a)_/m\\_.e@_[1]-', 'n a m e 1')
])
def test_clean_name(name, expected):
    assert common.clean_name(name) == expected


bad_network = site_dicts()[0]
bad_network['extra_parameters']['network'] = 'BBQ DISCO'
bad_network = expected_site(bad_network)


@pytest.mark.parametrize('networks,site,expected', [
    (['DOE ARM'], site_string_dicts[0], True),
    ('DOE ARM', site_string_dicts[0], True),
    (['DOE ARM', 'NREL MIDC'], site_string_dicts[1], False),
    (['NOAA SURFRAD', 'DOE ARM'], site_string_dicts[3], False),
    (['NOAA SURFRAD', 'DOE ARM'], bad_network, False),
    (['ARMS'], site_string_dicts[0], False),
])
def test_check_network(networks, site, expected):
    metadata = Site.from_dict(site)
    assert common.check_network(networks, metadata) == expected


@pytest.mark.parametrize('networks,expected', [
    (['DOE ARM'], site_objects[:1]),
    ('DOE ARM', site_objects[:1]),
    (['NOAA SURFRAD', 'NOAA SOLRAD'], site_objects[1:3])
])
def test_filter_by_network(networks, expected):
    assert common.filter_by_networks(site_objects, networks) == expected


@pytest.mark.parametrize('site,variable', [
    (site_objects[0], 'ghi'),
])
def test_create_observation(
        mock_api, site, variable, observation_objects_param):
    mock_api.list_observations.return_value = []
    common.create_observation(mock_api, site, variable)
    mock_api.create_observation.assert_called()
    mock_api.create_observation.assert_called_with(
        observation_objects_param[0])


def test_create_observation_exists(
        mock_api, site_objects_param, observation_objects_param):
    variable = 'ghi'
    site = site_objects_param[0]
    common.create_observation(mock_api, site, variable)
    mock_api.create_observation.assert_not_called()


long_site_name = Site.from_dict({
    'name': 'ARM site with just abouts sixty four characters in its name oops',
    'latitude': 1,
    'longitude': 1,
    'elevation': 5,
    'timezone': 'Etc/GMT+8',
    'extra_parameters': ('{"network": "DOE ARM", "network_api_abbreviation": '
                         '"site_abbrev", "network_api_id": "thing", '
                         '"observation_interval_length": 1}'),
})
observation_long_site_name = Observation.from_dict({
    'name': 'site_abbrev air_temperature',
    'variable': 'air_temperature',
    'interval_label': 'ending',
    'interval_value_type': 'interval_mean',
    'interval_length': 1,
    'site': long_site_name,
    'uncertainty': 0,
    'extra_parameters': ('{"network": "DOE ARM", "network_api_abbreviation": '
                         '"site_abbrev", "network_api_id": "thing", '
                         '"observation_interval_length": 1}'),
})


@pytest.mark.parametrize('site,variable,expected', [
    (long_site_name, 'air_temperature', observation_long_site_name),
])
def test_create_observation_long_site(mock_api, site, variable, expected):
    common.create_observation(mock_api, site, variable)
    mock_api.create_observation.assert_called()
    mock_api.create_observation.assert_called_with(expected)


observation_with_extra_params = Observation.from_dict({
    'name': 'site ghi',
    'variable': 'ghi',
    'interval_label': 'ending',
    'interval_value_type': 'interval_mean',
    'interval_length': 5,
    'site': site_objects[0],
    'uncertainty': 0,
    'extra_parameters': '{"network": "", "observation_interval_length": 5}'
})
observation_params = {
    'network': '',
    'observation_interval_length': 5,
}


@pytest.mark.parametrize('site,variable,expected,extra_params', [
    (site_objects[0], 'ghi', observation_with_extra_params,
     observation_params),
])
def test_create_observation_extra_parameters(
        mock_api, site, variable, expected, extra_params):
    mock_api.list_observations.return_value = []
    common.create_observation(mock_api, site, variable, extra_params)
    mock_api.create_observation.assert_called()
    mock_api.create_observation.assert_called_with(expected)


observation_dict = {
    'name': 'just observation',
    'variable': 'ghi',
    'interval_label': 'beginning',
    'interval_value_type': 'instantaneous',
    'interval_length': 1,
    'site': site_objects[0],
    'uncertainty': 2,
    'extra_parameters': ('{"network": "DOE ARM", '
                         '"network_api_id": "B13", '
                         '"network_api_abbreviation": "sgp", '
                         '"observation_interval_length": 1, '
                         '"datastreams": {'
                         '"qcrad": "sgpqcradlong1E13.c1", '
                         '"met": {'
                         '"sgpmetE13.b1": "2019-01-01/2020-01-01", '
                         '"sgpmetE13.b2": "2020-01-01/2021-01-01"}}}')
}
test_kwarg_observation = Observation.from_dict(observation_dict)
obs_kwargs = {
    'interval_label': 'beginning',
    'name': 'just observation',
    'interval_value_type': 'instantaneous',
    'uncertainty': 2,
}

observation_dict_resample_mean = observation_dict.copy()
observation_dict_resample_mean['extra_parameters'] = (
    '{"network": "DOE ARM", '
    '"network_api_id": "qcradlong1", '
    '"network_api_abbreviation": "abbrv", '
    '"observation_interval_length": 1, '
    '"resample_how": "mean"}')
test_observation_mean = Observation.from_dict(observation_dict_resample_mean)

observation_dict_resample_first = observation_dict.copy()
observation_dict_resample_first['extra_parameters'] = (
    '{"network": "DOE ARM", '
    '"network_api_id": "qcradlong1", '
    '"network_api_abbreviation": "abbrv", '
    '"observation_interval_length": 1, '
    '"resample_how": "first"}')
test_observation_first = Observation.from_dict(observation_dict_resample_first)

observation_dict_resample_fail = observation_dict.copy()
observation_dict_resample_fail['extra_parameters'] = (
    '{"network": "DOE ARM", '
    '"network_api_id": "qcradlong1", '
    '"network_api_abbreviation": "abbrv", '
    '"observation_interval_length": 1, '
    '"resample_how": "nopethepope"}')
test_observation_fail = Observation.from_dict(observation_dict_resample_fail)


@pytest.mark.parametrize('site,variable,expected,kwargs', [
    (site_objects[0], 'ghi', test_kwarg_observation, obs_kwargs),
])
def test_create_observation_with_kwargs(
        mock_api, site, variable, expected, kwargs):
    common.create_observation(mock_api, site, variable, **kwargs)
    mock_api.create_observation.assert_called()
    mock_api.create_observation.assert_called_with(expected)


@pytest.mark.parametrize('observation,resample_how', [
    # all test data has same freq as observation.interval_length,
    # so resample method should not matter
    (test_kwarg_observation, None),
    (test_observation_mean, 'mean'),
    (test_observation_first, 'first')
])
@pytest.mark.parametrize('inp,expected', [
    # nan kept
    (pd.DataFrame({'ghi': [0, 1, 4.0, None, 5, 6]}, dtype='float',
                  index=pd.date_range('2019-10-04T1200Z', freq='1min',
                                      periods=6)),
     pd.DataFrame({'value': [0, 1, 4.0, None, 5, 6],
                   'quality_flag': [0, 0, 0, 1, 0, 0]},
                  index=pd.date_range('2019-10-04T1200Z', freq='1min',
                                      periods=6))),
    # nan filled in center
    (pd.DataFrame({'ghi': [0, 1, 4.0, 5, 6]}, dtype='float',
                  index=pd.DatetimeIndex(
                      ['2019-10-04T1200Z', '2019-10-04T1201Z',
                       '2019-10-04T1202Z', '2019-10-04T1204Z',
                       '2019-10-04T1205Z'])),
     pd.DataFrame({'value': [0, 1, 4.0, None, 5, 6],
                   'quality_flag': [0, 0, 0, 1, 0, 0]},
                  index=pd.date_range('2019-10-04T1200Z', freq='1min',
                                      periods=6))),
    # new freq
    (pd.DataFrame({'ghi': [0, 1]}, dtype='float',
                  index=pd.DatetimeIndex(
                      ['2019-10-04T1200Z', '2019-10-04T1205Z'])),
     pd.DataFrame({'value': [0, None, None, None, None, 1.0],
                   'quality_flag': [0, 1, 1, 1, 1, 0]},
                  index=pd.date_range('2019-10-04T1200Z', freq='1min',
                                      periods=6))),

    # nans not extended to start end
    (pd.DataFrame({'ghi': [0, 1]}, dtype='float',
                  index=pd.DatetimeIndex(
                      ['2019-10-04T1201Z', '2019-10-04T1202Z'])),
     pd.DataFrame({'value': [0.0, 1.0],
                   'quality_flag': [0, 0]},
                  index=pd.DatetimeIndex(
                      ['2019-10-04T1201Z', '2019-10-04T1202Z']))),
])
def test_prepare_data_to_post(inp, expected, observation, resample_how):
    start = pd.Timestamp('2019-10-04T1200Z')
    end = pd.Timestamp('2019-10-04T1205Z')
    variable = 'ghi'
    out = common._prepare_data_to_post(inp, variable, observation, start, end,
                                       resample_how=resample_how)
    pd.testing.assert_frame_equal(out, expected)


@pytest.mark.parametrize('inp,expected', [
    (pd.DataFrame({'ghi': [0, 1, 4.0, None, 5, 6] * 2}, dtype='float',
                  index=pd.date_range('2019-10-04T1200Z', freq='30s',
                                      periods=12)),
     pd.DataFrame({'value': [0.5, 4.0, 5.5, 0.5, 4.0, 5.5],
                   'quality_flag': [0, 0, 0, 0, 0, 0]},
                  index=pd.date_range('2019-10-04T1200Z', freq='1min',
                                      periods=6))),
    (pd.DataFrame({'ghi': [0, 1, None, None, 5, 6] * 2}, dtype='float',
                  index=pd.date_range('2019-10-04T1200Z', freq='30s',
                                      periods=12)),
     pd.DataFrame({'value': [0.5, None, 5.5, 0.5, None, 5.5],
                   'quality_flag': [0, 1, 0, 0, 1, 0]},
                  index=pd.date_range('2019-10-04T1200Z', freq='1min',
                                      periods=6))),
])
def test_prepare_data_to_post_mean(inp, expected):
    start = pd.Timestamp('2019-10-04T1200Z')
    end = pd.Timestamp('2019-10-04T1205Z')
    variable = 'ghi'
    out = common._prepare_data_to_post(inp, variable, test_observation_mean,
                                       start, end, resample_how='mean')
    pd.testing.assert_frame_equal(out, expected)


@pytest.mark.parametrize('inp,expected', [
    (pd.DataFrame({'ghi': [0, 1, None, 4.0, 5, 6] * 2}, dtype='float',
                  index=pd.date_range('2019-10-04T1200Z', freq='30s',
                                      periods=12)),
     pd.DataFrame({'value': [0.0, 4.0, 5.0, 0.0, 4.0, 5.0],
                   'quality_flag': [0, 0, 0, 0, 0, 0]},
                  index=pd.date_range('2019-10-04T1200Z', freq='1min',
                                      periods=6))),
    (pd.DataFrame({'ghi': [0, 1, None, None, 5, 6] * 2}, dtype='float',
                  index=pd.date_range('2019-10-04T1200Z', freq='30s',
                                      periods=12)),
     pd.DataFrame({'value': [0.0, None, 5.0, 0.0, None, 5.0],
                   'quality_flag': [0, 1, 0, 0, 1, 0]},
                  index=pd.date_range('2019-10-04T1200Z', freq='1min',
                                      periods=6))),
])
def test_prepare_data_to_post_first(inp, expected):
    start = pd.Timestamp('2019-10-04T1200Z')
    end = pd.Timestamp('2019-10-04T1205Z')
    variable = 'ghi'
    out = common._prepare_data_to_post(inp, variable, test_observation_first,
                                       start, end, resample_how='first')
    pd.testing.assert_frame_equal(out, expected)


def test_prepare_data_to_post_offset():
    # offset data is kept as is. If this is a problem we'll need to
    # probably resample the reference data to the appropriate index
    start = pd.Timestamp('2019-10-04T1200Z')
    end = pd.Timestamp('2019-10-04T1215Z')
    variable = 'ghi'
    inp = pd.DataFrame({'ghi': [0, 1, 4.0]},
                       index=pd.date_range('2019-10-04T1201Z', freq='5min',
                                           periods=3))
    expected = pd.DataFrame(
        {'value': [0, 1, 4.0], 'quality_flag': [0, 0, 0]},
        index=pd.date_range('2019-10-04T1201Z', freq='5min',
                            periods=3))
    out = common._prepare_data_to_post(
        inp, variable, observation_with_extra_params, start, end)
    pd.testing.assert_frame_equal(out, expected)


def test_prepare_data_to_post_empty():
    inp = pd.DataFrame(
        {'ghi': [0.0]}, index=pd.DatetimeIndex(['2019-01-01 00:00Z']))
    start = pd.Timestamp('2019-10-04T1200Z')
    end = pd.Timestamp('2019-10-04T1205Z')
    variable = 'ghi'
    out = common._prepare_data_to_post(inp, variable, test_kwarg_observation,
                                       start, end)
    assert out.empty


def test_prepare_data_to_post_no_var():
    start = pd.Timestamp('2109-10-04T1200Z')
    end = pd.Timestamp('2109-10-04T1300Z')
    data = pd.DataFrame({'notavar': [0, 1]})
    with pytest.raises(KeyError):
        common._prepare_data_to_post(data, 'GHI 7', test_kwarg_observation,
                                     start, end)


def test_post_observation_data_no_data(mock_api, log, start, end,):
    common.post_observation_data(
        mock_api, test_kwarg_observation,
        pd.DataFrame({'a': [1, 2, 3],
                      'b': ['a', 'b', 'c']}),
        start, end)
    log.error.assert_called()


def test_post_observation_data(mock_api, log, fake_ghi_data, start, end):
    common.post_observation_data(mock_api, site_test_observation,
                                 fake_ghi_data, start, end)

    args, _ = mock_api.post_observation_values.call_args
    # test observations never get assigned an id so the observation_id
    # argument should be an empty string
    assert args[0] == ''
    pd.testing.assert_frame_equal(
        args[1], fake_ghi_data.rename(columns={'ghi': 'value'}))


def test_post_observation_data_HTTPError(
        mock_api, log, fake_ghi_data, start, end):
    # mocking this error means that when the debugging logging call
    # fires, an AttributeError is thrown while looking for a response
    # but this also tests that we've followed the correct logic.
    mock_api.post_observation_values.side_effect = HTTPError
    with pytest.raises(AttributeError):
        common.post_observation_data(mock_api, site_test_observation,
                                     fake_ghi_data, start, end)
    log.error.assert_called()


def test_post_observation_data_all_nans(
        mock_api, log, fake_ghi_data, start, end):
    nan_data = fake_ghi_data.copy()
    nan_data['ghi'] = np.NaN
    ret = common.post_observation_data(mock_api, site_test_observation,
                                       nan_data, start, end)
    # post the nans
    log.warning.assert_not_called()
    assert ret is None


def test_post_observation_data_AttributeError(
        mock_api, log, fake_ghi_data, start, end):
    common.post_observation_data(mock_api, test_observation_fail,
                                 fake_ghi_data, start, end)
    log.error.assert_called()


@pytest.fixture()
def now(mocker):
    now = pd.Timestamp('20190108T11:32:00Z')
    mocker.patch(
        'solarforecastarbiter.io.reference_observations.common._utcnow',
        return_value=now)
    return now


@pytest.fixture()
def site_obs():
    return [
        Observation.from_dict({
            'name': 'site ghi',
            'observation_id': str(uuid.uuid1()),
            'variable': 'ghi',
            'interval_label': 'ending',
            'interval_value_type': 'interval_mean',
            'interval_length': '5',
            'site': site,
            'uncertainty': 0,
            'extra_parameters': site.extra_parameters
        }) for site in site_objects[:2]]


def test_get_last_site_timestamp(mock_api, now, site_obs):
    retvals = {site_obs[0].observation_id: (0, pd.Timestamp('20190105T0020Z')),
               site_obs[1].observation_id: (0, now)}
    mock_api.get_observation_time_range.side_effect = lambda x: retvals[x]
    ret = common.get_last_site_timestamp(mock_api, site_obs, now)
    assert ret == pd.Timestamp('20190105T0020Z')


def test_get_last_site_timestamp_small(mock_api, now, site_obs):
    retvals = {site_obs[0].observation_id: (0, pd.Timestamp('20181222T0020Z')),
               site_obs[1].observation_id: (0, pd.Timestamp('20190106T0020Z'))}
    mock_api.get_observation_time_range.side_effect = lambda x: retvals[x]
    ret = common.get_last_site_timestamp(mock_api, site_obs, now)
    assert ret == now - pd.Timedelta('7d')


def test_get_last_site_timestamp_none(mock_api, now, site_obs):
    mock_api.get_observation_time_range.return_value = (0, pd.NaT)
    ret = common.get_last_site_timestamp(mock_api, site_obs, now)
    assert ret == now - pd.Timedelta('7d')


def test_get_last_site_timestamp_some_nat(mock_api, now, site_obs):
    retvals = {site_obs[0].observation_id: (0, pd.Timestamp('20190105T0020Z')),
               site_obs[1].observation_id: (0, pd.NaT)}
    mock_api.get_observation_time_range.side_effect = lambda x: retvals[x]
    ret = common.get_last_site_timestamp(mock_api, site_obs, now)
    assert ret == pd.Timestamp('20190105T0020Z')


def test_get_last_site_timestamp_uptodate(mock_api, now, site_obs):
    mock_api.get_observation_time_range.return_value = (0, now)
    ret = common.get_last_site_timestamp(mock_api, site_obs, now)
    assert ret == now


def test_get_last_site_timestamp_empty(mock_api, now):
    ret = common.get_last_site_timestamp(mock_api, [], now)
    assert ret == now - pd.Timedelta('7d')


@pytest.mark.parametrize('site', site_objects)
def test_update_site_observations(
        mock_api, mock_fetch, site, observation_objects_param, fake_ghi_data):
    start = pd.Timestamp('20190101T1205Z')
    end = pd.Timestamp('20190101T1225Z')
    common.update_site_observations(
        mock_api, mock_fetch, site, observation_objects_param, start, end)
    args, _ = mock_api.post_observation_values.call_args
    assert args[0] == ''
    pd.testing.assert_frame_equal(
        args[1], fake_ghi_data.rename(
            columns={'ghi': 'value'})[start:end].resample(
                args[1].index.freq).first())


def test_update_site_observations_no_start(
        mock_api, mock_fetch, observation_objects_param, fake_ghi_data,
        now):
    site = site_objects[0]
    start = None
    slimit = pd.Timestamp('20190101T1219Z')
    end = pd.Timestamp('20190108T1219Z')
    mock_api.get_observation_time_range.return_value = (0, slimit)
    common.update_site_observations(
        mock_api, mock_fetch, site, observation_objects_param, start, end)
    args, _ = mock_api.post_observation_values.call_args
    assert args[0] == ''
    pd.testing.assert_frame_equal(
        args[1], fake_ghi_data.rename(
            columns={'ghi': 'value'})[slimit:end].resample(
                args[1].index.freq).first())


def test_update_site_observations_no_end(
        mock_api, mock_fetch, observation_objects_param, fake_ghi_data,
        now):
    site = site_objects[0]
    start = pd.Timestamp('20190101T1210Z')
    end = None
    common.update_site_observations(
        mock_api, mock_fetch, site, observation_objects_param, start, end)
    args, _ = mock_api.post_observation_values.call_args
    assert args[0] == ''
    pd.testing.assert_frame_equal(
        args[1], fake_ghi_data.rename(
            columns={'ghi': 'value'})[start:].resample(
                args[1].index.freq).first())


def test_update_site_observations_uptodate(
        mock_api, mock_fetch, observation_objects_param, fake_ghi_data,
        now):
    site = site_objects[0]
    mock_api.get_observation_time_range.return_value = (0, now)
    common.update_site_observations(
        mock_api, mock_fetch, site, observation_objects_param, None, None)
    mock_api.post_observation_values.assert_not_called


def test_update_site_observations_no_data(
        mock_api, mocker, site_objects_param,
        observation_objects_param, log, start, end):
    fetch = mocker.MagicMock()
    fetch.return_value = pd.DataFrame()
    common.update_site_observations(
        mock_api, fetch, site_objects[1],
        observation_objects_param, start, end)
    mock_api.assert_not_called()


def test_update_site_observations_out_of_order(
        mock_api, site_objects_param, mocker,
        observation_objects_param, fake_ghi_data):
    start = pd.Timestamp('20190101T1200Z')
    end = pd.Timestamp('20190101T1230Z')
    fetch = mocker.MagicMock()
    fetch.return_value = fake_ghi_data.sample(frac=1)
    common.update_site_observations(
        mock_api, fetch, site_objects[1], observation_objects_param,
        start, end)
    args, _ = mock_api.post_observation_values.call_args
    assert args[0] == ''
    pd.testing.assert_frame_equal(
        args[1], fake_ghi_data.rename(
            columns={'ghi': 'value'})[start:end].resample(
                args[1].index.freq).first())


@pytest.fixture()
def template_fx(mock_api, mocker):
    mock_api.create_forecast = mocker.MagicMock(side_effect=lambda x: x)
    mock_api.create_probabilistic_forecast = mocker.MagicMock(
        side_effect=lambda x: x)
    site = site_objects[1].replace(latitude=32, longitude=-110)
    template = Forecast(
        name='Test Template',
        issue_time_of_day=dt.time(0),
        lead_time_to_start=pd.Timedelta('1d'),
        interval_length=pd.Timedelta('1h'),
        run_length=pd.Timedelta('24h'),
        interval_label='ending',
        interval_value_type='interval_mean',
        variable='ghi',
        site=Site('dummy', 0, 0, 0, 'MST'),
        extra_parameters=json.dumps(
            {'is_reference_forecast': True,
             'model': 'gfs_quarter_deg_hourly_to_hourly_mean'
             })
        )
    return mock_api, template, site


def test_create_one_forecast(template_fx):
    api, template, site = template_fx
    fx = common.create_one_forecast(api, site, template, 'ac_power')
    assert fx.name == 'site2 Test Template ac_power'
    assert fx.variable == 'ac_power'
    assert fx.site == site
    assert fx.issue_time_of_day == dt.time(8)
    ep = json.loads(fx.extra_parameters)
    assert ep['is_reference_forecast']
    assert ep['model'] == 'gfs_quarter_deg_hourly_to_hourly_mean'
    assert 'piggyback_on' not in ep


def test_create_one_forecast_invalid(template_fx):
    api, template, site = template_fx

    def fail(fx):
        raise ValueError('failed')
    fx = common.create_one_forecast(
        api, site, template, 'ac_power',
        creation_validation=fail)
    assert fx is None


@pytest.mark.parametrize('tz,expected', [
    ('Etc/GMT+8', dt.time(1)),
    ('MST', dt.time(0)),
    ('Etc/GMT+5', dt.time(4))
])
def test_create_one_forecast_issue_time(template_fx, tz, expected):
    api, template, site = template_fx
    template = template.replace(run_length=pd.Timedelta('6h'),
                                issue_time_of_day=dt.time(5),
                                lead_time_to_start=pd.Timedelta('1h'),
                                )
    site = site.replace(timezone=tz)
    fx = common.create_one_forecast(api, site, template, 'ac_power')
    assert fx.issue_time_of_day == expected


def test_create_one_forecast_long_name(template_fx):
    api, template, site = template_fx
    nn = 'a ' + ''.join(['n'] * 64)
    fx = common.create_one_forecast(api, site.replace(name=nn), template,
                                    'ac_power')
    assert fx.name == 'a Test Template ac_power'
    assert fx.variable == 'ac_power'
    assert fx.issue_time_of_day == dt.time(8)
    ep = json.loads(fx.extra_parameters)
    assert ep['is_reference_forecast']
    assert ep['model'] == 'gfs_quarter_deg_hourly_to_hourly_mean'
    assert 'piggyback_on' not in ep


@pytest.mark.parametrize('tmpl,exp', [
    ('Persistence 1hour ahead', 'The Site is really really really really Pers 1hour ahead GHI'),  # NOQA
    ('Persistence Fifteen-minute ahead', 'The Site is really really really really Pers 15 min ahead GHI'),  # NOQA
    ('n' * 46, 'The Site is ' + 'n' * 46 + ' GHI'),
    pytest.param(
        'This is too long to be a template forecast name and will raise a value error',   # NOQA
        '', marks=pytest.mark.xfail(strict=True, raises=ValueError))
])
def test_make_fx_name(tmpl, exp):
    out = common._make_fx_name(
        'The Site is really really really really long', tmpl, 'GHI')
    assert out == exp


def test_make_fx_name_long_site():
    with pytest.raises(ValueError):
        common._make_fx_name(
            'a' * 14,
            'r' * 49,
            '')


def test_create_one_forecast_piggy(template_fx):
    api, template, site = template_fx
    fx = common.create_one_forecast(api, site, template, 'ac_power',
                                    piggyback_on='other_fx')
    assert fx.name == 'site2 Test Template ac_power'
    assert fx.variable == 'ac_power'
    assert fx.site == site
    assert fx.issue_time_of_day == dt.time(8)
    ep = json.loads(fx.extra_parameters)
    assert ep['is_reference_forecast']
    assert ep['model'] == 'gfs_quarter_deg_hourly_to_hourly_mean'
    assert ep['piggyback_on'] == 'other_fx'


def test_create_one_forecast_existing(template_fx, mocker):
    api, template, site = template_fx
    newfx = template.replace(name='site2 Test Template ac_power', site=site)
    api.list_forecasts = mocker.MagicMock(return_value=[newfx])
    fx = common.create_one_forecast(api, site, template, 'ac_power',
                                    piggyback_on='other_fx')
    assert fx == newfx


@pytest.mark.parametrize('vars_,primary', [
    (('ac_power', 'dni'), 'ac_power'),
    (('ghi', 'dni'), 'ghi'),
    (('dni', 'wind_speed', 'relative_humidity'), False)
])
def test_create_nwp_forecasts(template_fx, mocker, vars_, primary):
    api, template, site = template_fx
    templates = [template.replace(name='one'), template.replace(name='two')]
    fxdict = template.to_dict()
    fxdict['constant_values'] = [0, 50, 100]
    fxdict['axis'] = 'y'
    templates += [ProbabilisticForecast.from_dict(fxdict)]

    fxs = common.create_nwp_forecasts(api, site, vars_, templates)
    assert len(fxs) == 6
    if primary:
        assert fxs[0].variable == primary
        assert fxs[2].variable == primary
    assert 'one' in fxs[0].name
    assert 'one' in fxs[1].name
    assert fxs[0].forecast_id in fxs[1].extra_parameters
    assert 'two' in fxs[2].name
    assert 'two' in fxs[3].name
    assert fxs[2].forecast_id in fxs[3].extra_parameters
    assert isinstance(fxs[-1], ProbabilisticForecast)


def test_create_nwp_forecasts_outside(template_fx, mocker, log):
    vars_ = ('ac_power', 'dni')
    api, template, site = template_fx
    site = site.replace(latitude=19, longitude=-159)
    templates = [template.replace(name='one'), template.replace(name='two')]
    with pytest.raises(ValueError):
        common.create_nwp_forecasts(api, site, vars_, templates)


def test_create_persistence_forecasts(template_fx, mocker):
    vars_ = ('ghi', 'dni', 'air_temperature', 'ac_power')
    api, template, site = template_fx
    templates = [
        template.replace(extra_parameters=json.dumps(
            {'is_reference_persistence_forecast': True}),
                         run_length=pd.Timedelta('6h')),
        template.replace(extra_parameters=json.dumps(
            {'is_reference_persistence_forecast': True}),
                         name='Longer template')
    ]
    obss = [site_test_observation.replace(site=site),
            site_test_observation.replace(site=site, variable='dni'),
            site_test_observation,
            site_test_observation.replace(
                site=site, variable='air_temperature'),
            site_test_observation.replace(
                site=site, variable='dhi'),
            ]
    api.list_observations = mocker.MagicMock(return_value=obss)
    fxs = common.create_persistence_forecasts(api, site, vars_, templates)
    assert len(fxs) == 6
    assert sorted([fx.variable for fx in fxs]) == [
        'air_temperature', 'air_temperature', 'dni', 'dni', 'ghi', 'ghi']
    assert 'Longer' in fxs[-1].name
    assert all(['observation_id' in fx.extra_parameters for fx in fxs])
    index_fxs = [re.search(r'index_persistence(["\s\:]*)true',
                           fx.extra_parameters, re.I) is not None
                 for fx in fxs if fx.variable != 'air_temperature'
                 and 'Longer' not in fx.name
                 ]
    assert len(index_fxs) == 2
    assert all(index_fxs)


def test_create_forecasts(template_fx, mocker):
    vars_ = ('ac_power', 'ghi')
    api, template, site = template_fx
    templates = [template, template.replace(extra_parameters=json.dumps(
        {'is_reference_persistence_forecast': True}))]
    api.list_observations = mocker.MagicMock(
        return_value=[site_test_observation.replace(site=site)])
    create_nwp = mocker.spy(common, 'create_nwp_forecasts')
    create_perst = mocker.spy(common, 'create_persistence_forecasts')
    fxs = common.create_forecasts(api, site, vars_, templates)
    assert len(fxs) == 3
    assert create_nwp.call_count == 1
    assert create_perst.call_count == 1
    assert create_nwp.call_args[0][-1] == templates[:1]
    assert create_perst.call_args[0][-1] == templates[-1:]


@pytest.mark.parametrize('params', [
    {'network_api_id': 2},
    {'network_api_id': '2'},
])
def test_apply_json_site_parameters_plant(test_json_site_file, params):
    new_site = common.apply_json_site_parameters(
        test_json_site_file,
        {'extra_parameters': params},
    )
    assert 'modeling_parameters' in new_site
    extra_params = new_site['extra_parameters']
    assert extra_params['network_api_abbreviation'] == 'SITE2'
    assert extra_params['attribution'] == ""
    assert extra_params['network'] == 'TEST'
    assert extra_params['observation_interval_length'] == 1.0


@pytest.mark.parametrize('params', [
    {'network_api_id': 'not_plant'},
])
def test_apply_json_site_parameters_no_params(test_json_site_file, params):
    new_site = common.apply_json_site_parameters(
        test_json_site_file,
        {'extra_parameters': params},
    )
    assert 'modeling_parameters' not in new_site
    assert list(new_site['extra_parameters'].keys()) == ['network_api_id']



"""Probablistic forecast error metrics."""

import numpy as np


def brier_score(obs, fx, fx_prob):
    """Brier Score (BS).

        BS = 1/n sum_{i=1}^n (f_i - o_i)^2

    where n is the number of forecasts, f_i is the forecasted probability of
    event i, and o_i is the observed event indicator (o_i=0: event did not
    occur, o_i=1: event occured). The forecasts are supplied as the
    right-hand-side of a CDF interval, e.g., forecast <= 10 MW at time i, and
    therefore o_i is defined as:

        o_i = 1 if obs_i <= fx_i, else o_i = 0

    where fx_i and obs_i are the forecast and observation at time i,
    respectively.

    Parameters
    ----------
    obs : (n,) array_like
        Observations (physical unit).
    fx : (n,) array_like
        Forecasts (physical units) of the right-hand-side of a CDF interval,
        e.g., fx = 10 MW is interpreted as forecasting <= 10 MW.
    fx_prob : (n,) array_like
        Probability [%] associated with the forecasts.

    Returns
    -------
    bs : float
        The Brier Score [unitless], bounded between 0 and 1, where values
        closer to 0 indicate better forecast performance and values closer to 1
        indicate worse performance.

    Notes
    -----
    The Brier Score implemented in this function is for binary outcomes only,
    rather than the more general (but less commonly used) categorical version.

    """

    # event: 0=did not happen, 1=did happen
    o = np.where(obs <= fx, 1.0, 0.0)

    # forecast probabilities [unitless]
    f = fx_prob / 100.0

    bs = np.mean((f - o) ** 2)
    return bs


def brier_skill_score(obs, fx, fx_prob, ref, ref_prob):
    """Brier Skill Score (BSS).

        BSS = 1 - BS_fx / BS_ref

    where BS_fx is the Brier Score of the evaluated forecast and BS_ref is the
    Brier Score of a reference forecast.

    Parameters
    ----------
    obs : (n,) array_like
        Observations (physical unit).
    fx : (n,) array_like
        Forecasts (physical units) of the right-hand-side of a CDF interval,
        e.g., fx = 10 MW is interpreted as forecasting <= 10 MW.
    fx_prob : (n,) array_like
        Probability [%] associated with the forecasts.
    ref : (n,) array_like
        Reference forecast (physical units) of the right-hand-side of a CDF
        interval.
    ref_prob : (n,) array_like
        Probability [%] associated with the reference forecast.

    Returns
    -------
    skill : float
        The Brier Skill Score [unitless].

    """
    bs_fx = brier_score(obs, fx, fx_prob)
    bs_ref = brier_score(obs, ref, ref_prob)
    skill = 1.0 - bs_fx / bs_ref
    return skill


def quantile_score(obs, fx, fx_prob):
    """Quantile Score (QS).

    .. math::

        \\text{QS} = \\frac{1}{n} \\sum_{i=1}^n (fx_i - obs_i) * (p - 1\\{obs_i > fx_i\\})

    where :math:`n` is the number of forecasts, :math:`obs_i` is an
    observation, :math:`fx_i` is a forecast, :math:`1\\{obs_i > fx_i\\}` is an
    indicator function (1 if :math:`obs_i > fx_i`, 0 otherwise) and :math:`p`
    is the probability that :math:`obs_i <= fx_i`. [1]_ [2]_

    If :math:`obs > fx`, then we have:

    .. math::

        (fx - obs) < 0 \\\\
        (p - 1\\{obs > fx\\}) = (p - 1) <= 0 \\\\
        (fx - obs) * (p - 1) >= 0

    If instead :math:`obs < fx`, then we have:

    .. math::

        (fx - obs) > 0 \\\\
        (p - 1\\{obs > fx\\}) = (p - 0) >= 0 \\\\
        (fx - obs) * p >= 0

    Therefore, the quantile score is non-negative regardless of the obs and fx.

    Parameters
    ----------
    obs : (n,) array_like
        Observations (physical unit).
    fx : (n,) array_like
        Forecasts (physical units) of the right-hand-side of a CDF interval,
        e.g., fx = 10 MW is interpreted as forecasting <= 10 MW.
    fx_prob : (n,) array_like
        Probability [%] associated with the forecasts.

    Returns
    -------
    qs : float
        The Quantile Score, with the same units as the observations.

    Notes
    -----
    Quantile score is meant to be computed for a single probability of
    :math:`n` samples.

    Examples
    --------
    >>> obs = 100     # observation [MW]
    >>> fx = 80       # forecast [MW]
    >>> fx_prob = 60  # probability [%]
    >>> quantile_score(obs, fx, fx_prob)   # score [MW]
    8.0

    References
    ----------
    .. [1] Koenker and Bassett, Jr. (1978) "Regression Quantiles", Econometrica
       46 (1), pp. 33-50. doi: 10.2307/1913643
    .. [2] Wilks (2020) "Forecast Verification". In "Statistical Methods in the
       Atmospheric Sciences" (3rd edition). Academic Press. ISBN: 9780123850225

    """  # NOQA: E501,W605

    # Prob(obs <= fx) = p
    p = fx_prob / 100.0
    qs = np.mean((fx - obs) * (p - np.where(obs > fx, 1.0, 0.0)))
    return qs


def quantile_skill_score(obs, fx, fx_prob, ref, ref_prob):
    """Quantile Skill Score (QSS).

    .. math::

        \\text{QSS} = 1 - \\text{QS}_{\\text{fx}} / \\text{QS}_{\\text{ref}}

    where :math:`\\text{QS}_{\\text{fx}}` is the Quantile Score of the
    evaluated forecast and :math:`\\text{QS}_{\\text{ref}}` is the Quantile
    Score of a reference forecast. [1]_

    Parameters
    ----------
    obs : (n,) array_like
        Observations (physical unit).
    fx : (n,) array_like
        Forecasts (physical units) of the right-hand-side of a CDF interval,
        e.g., fx = 10 MW is interpreted as forecasting <= 10 MW.
    fx_prob : (n,) array_like
        Probability [%] associated with the forecasts.
    ref : (n,) array_like
        Reference forecast (physical units) of the right-hand-side of a CDF
        interval.
    ref_prob : (n,) array_like
        Probability [%] associated with the reference forecast.

    Returns
    -------
    skill : float
        The Quantile Skill Score [unitless].

    References
    ----------
    .. [1] Bouallegue, Pinson and Friederichs (2015) "Quantile forecast
       discrimination ability and value", Quarterly Journal of the Royal
       Meteorological Society 141, pp. 3415-3424. doi: 10.1002/qj.2624

    Notes
    -----
    This function returns 0 if QS_fx and QS_ref are both 0.

    See Also
    --------
    :py:func:`solarforecastarbiter.metrics.probabilistic.quantile_score`

    """

    qs_fx = quantile_score(obs, fx, fx_prob)
    qs_ref = quantile_score(obs, ref, ref_prob)

    # avoid 0 / 0 --> nan
    if qs_fx == qs_ref:
        return 0.0
    elif qs_ref == 0.0:
        # avoid divide by 0
        # typically caused by deadbands and short time periods
        return np.NINF
    else:
        return 1.0 - qs_fx / qs_ref


def _unique_forecasts(f):
    """Convert forecast probabilities to a set of unique values.

    Determine a set of unique forecast probabilities, based on input forecast
    probabilities of arbitrary precision, and approximate the input
    probabilities to lie within the set of unique values.

    Parameters
    ----------
    f : (n,) array_like
        Probability [unitless] associated with the forecasts.

    Returns
    -------
    f_uniq : (n,) array_like
        The converted forecast probabilities [unitless].

    Notes
    -----
    This implementation determines the set of unique forecast probabilities by
    rounding the input probabilities to a precision determined by the number of
    input probability values: if less than 1000 samples, bin by tenths;
    otherwise bin by hundredths.

    Examples
    --------
    >>> f = np.array([0.1234, 0.156891, 0.10561])
    >>> _unique_forecasts(f)
    array([0.1, 0.2, 0.1])

    """

    if len(f) >= 1000:
        n_decimals = 2  # bin by hundredths (0.01, 0.02, etc.)
    else:
        n_decimals = 1  # bin by tenths (0.1, 0.2, etc.)

    f_uniq = np.around(f, decimals=n_decimals)
    return f_uniq


def brier_decomposition(obs, fx, fx_prob):
    """The 3-component decomposition of the Brier Score.

        BS = REL - RES + UNC

    where REL is the reliability, RES is the resolution and UNC is the
    uncertatinty.

    Parameters
    ----------
    obs : (n,) array_like
        Observations (physical unit).
    fx : (n,) array_like
        Forecasts (physical units) of the right-hand-side of a CDF interval,
        e.g., fx = 10 MW is interpreted as forecasting <= 10 MW.
    fx_prob : (n,) array_like
        Probability [%] associated with the forecasts.

    Returns
    -------
    rel : float
        The reliability of the forecast [unitless], where a perfectly reliable
        forecast has value of 0.
    res : float
        The resolution of the forecast [unitless], where higher values are
        better.
    unc : float
        The uncertainty [unitless], where lower values indicate the event being
        forecasted occurs rarely.

    Notes
    -----
    The current implementation iterates over the unique forecasts to compute
    the reliability and resolution, rather than using a vectorized formulation.
    While a vectorized formulation may be more computationally efficient, the
    clarity of the iterate version outweighs the efficiency gains from the
    vectorized version. Additionally, the number of unique forecasts is
    currently capped at 100, which small enough that there is likely no
    practical difference in computation time between the iterate vs vectorized
    versions.

    """

    # event: 0=did not happen, 1=did happen
    o = np.where(obs <= fx, 1.0, 0.0)

    # forecast probabilities [unitless]
    f = fx_prob / 100.0

    # get unique forecast probabilities by binning
    f = _unique_forecasts(f)

    # reliability and resolution
    rel, res = 0.0, 0.0
    o_avg = np.mean(o)
    for f_i, N_i in np.nditer(np.unique(f, return_counts=True)):
        o_i = np.mean(o[f == f_i])      # mean event value per set
        rel += N_i * (f_i - o_i) ** 2
        res += N_i * (o_i - o_avg) ** 2
    rel /= len(f)
    res /= len(f)

    # uncertainty
    base_rate = np.mean(o)
    unc = base_rate * (1.0 - base_rate)

    return rel, res, unc


def reliability(obs, fx, fx_prob):
    """Reliability (REL) of the forecast.

        REL = 1/n sum_{i=1}^I N_i (f_i - o_{i,avg})^2

    where n is the total number of forecasts, I is the number of unique
    forecasts (f_1, f_2, ..., f_I), N_i is the number of times each unique
    forecast occurs, o_{i,avg} is the average of the observed events during
    which the forecast was f_i.

    Parameters
    ----------
    obs : (n,) array_like
        Observations (physical unit).
    fx : (n,) array_like
        Forecasts (physical units) of the right-hand-side of a CDF interval,
        e.g., fx = 10 MW is interpreted as forecasting <= 10 MW.
    fx_prob : (n,) array_like
        Probability [%] associated with the forecasts.

    Returns
    -------
    rel : float
        The reliability of the forecast [unitless], where a perfectly reliable
        forecast has value of 0.

    See Also
    --------
    brier_decomposition : 3-component decomposition of the Brier Score

    """

    rel = brier_decomposition(obs, fx, fx_prob)[0]
    return rel


def resolution(obs, fx, fx_prob):
    """Resolution (RES) of the forecast.

        RES = 1/n sum_{i=1}^I N_i (o_{i,avg} - o_{avg})^2

    where n is the total number of forecasts, I is the number of unique
    forecasts (f_1, f_2, ..., f_I), N_i is the number of times each unique
    forecast occurs, o_{i,avg} is the average of the observed events during
    which the forecast was f_i, and o_{avg} is the average of all observed
    events.

    Parameters
    ----------
    obs : (n,) array_like
        Observations (physical unit).
    fx : (n,) array_like
        Forecasts (physical units) of the right-hand-side of a CDF interval,
        e.g., fx = 10 MW is interpreted as forecasting <= 10 MW.
    fx_prob : (n,) array_like
        Probability [%] associated with the forecasts.

    Returns
    -------
    res : float
        The resolution of the forecast [unitless], where higher values are
        better.

    See Also
    --------
    brier_decomposition : 3-component decomposition of the Brier Score

    """

    res = brier_decomposition(obs, fx, fx_prob)[1]
    return res


def uncertainty(obs, fx, fx_prob):
    """Uncertainty (UNC) of the forecast.

        UNC = base_rate * (1 - base_rate)

    where base_rate = 1/n sum_{i=1}^n o_i, and o_i is the observed event.

    Parameters
    ----------
    obs : (n,) array_like
        Observations (physical unit).
    fx : (n,) array_like
        Forecasts (physical units) of the right-hand-side of a CDF interval,
        e.g., fx = 10 MW is interpreted as forecasting <= 10 MW.
    fx_prob : (n,) array_like
        Probability [%] associated with the forecasts.

    Returns
    -------
    unc : float
        The uncertainty [unitless], where lower values indicate the event being
        forecasted occurs rarely.

    See Also
    --------
    brier_decomposition : 3-component decomposition of the Brier Score

    """

    unc = brier_decomposition(obs, fx, fx_prob)[2]
    return unc


def sharpness(fx_lower, fx_upper):
    """Sharpness (SH).

        SH = 1/n sum_{i=1}^n (f_{u,i} - f_{l,i})

    where n is the total number of forecasts, f_{u,i} is the upper prediction
    interval value and f_{l,i} is the lower prediction interval value for
    sample i.

    Parameters
    ----------
    fx_lower : (n,) array_like
        The lower prediction interval values (physical units).
    fx_upper : (n,) array_like
        The upper prediction interval values (physical units).

    Returns
    -------
    SH : float
        The sharpness (physical units), where smaller sharpness values indicate
        "tighter" prediction intervals.

    """
    sh = np.mean(fx_upper - fx_lower)
    return sh


def continuous_ranked_probability_score(obs, fx, fx_prob):
    """Continuous Ranked Probability Score (CRPS).

        CRPS = 1/n sum_{i=1}^n int (F_i - O_i)^2 dx

    where F_i is the CDF of the forecast at time i and O_i is the CDF
    associated with the observed value obs_i:

        O_{i, j} = 1 if obs_i <= fx_{i, j}, else O_{i, j} = 0

    where obs_i is the observation at time i, and fx_{i, j} is the forecast at
    time i for CDF interval j.

    Parameters
    ----------
    obs : (n,) array_like
        Observations (physical unit).
    fx : (n, d) array_like
        Forecasts (physical units) of the right-hand-side of a CDF with d
        intervals (d >= 2), e.g., fx = [10 MW, 20 MW, 30 MW] is interpreted as
        <= 10 MW, <= 20 MW, <= 30 MW.
    fx_prob : (n, d) array_like
        Probability [%] associated with the forecasts.

    Returns
    -------
    crps : float
        The Continuous Ranked Probability Score [unitless].

    Examples
    --------

    Forecast probabilities of <= 10 MW and <= 20 MW:
    >>> fx = np.array([[10, 20], [10, 20]])
    >>> fx_prob = np.array([[30, 50], [65, 100]])
    >>> obs = np.array([8, 12])
    >>> continuous_ranked_probability_score(obs, fx, fx_prob)
    4.5625

    Forecast thresholds for constant probabilities (25%, 75%):
    >>> fx = np.array([[5, 15], [8, 14]])
    >>> fx_prob = np.array([[25, 75], [25, 75]])
    >>> obs = np.array([8, 10])
    >>> continuous_ranked_probability_score(obs, fx, fx_prob)
    0.5

    """

    # match observations to fx shape: (n,) => (n, d)
    obs = np.tile(obs, (fx.shape[1], 1)).T

    # event: 0=did not happen, 1=did happen
    o = np.where(obs <= fx, 1.0, 0.0)

    # forecast probabilities [unitless]
    f = fx_prob / 100.0

    # integrate along each sample, then average all samples
    integrand = (f - o) ** 2
    dx = np.diff(fx, axis=1)
    crps = np.mean(np.sum(integrand[:, :-1] * dx, axis=1))
    return crps


# Add new metrics to this map to map shorthand to function
_MAP = {
    'bs': (brier_score, 'BS'),
    'bss': (brier_skill_score, 'BSS'),
    'rel': (reliability, 'REL'),
    'res': (resolution, 'RES'),
    'unc': (uncertainty, 'UNC'),
    'qs': (quantile_score, 'QS'),
    'qss': (quantile_skill_score, 'QSS'),
    # 'sh': (sharpness, 'SH'),  # TODO
    'crps': (continuous_ranked_probability_score, 'CRPS'),
}

__all__ = [m[0].__name__ for m in _MAP.values()]

# Functions that require a reference forecast
_REQ_REF_FX = ['bss', 'qss']

# Functions that require normalized factor
_REQ_NORM = []

# Functions that require full distribution forecasts (as 2dim)
_REQ_DIST = ['crps']

# TODO: Functions that require two forecasts (e.g., sharpness)
# _REQ_FX_FX = ['sh']
